/*****************************************************************************/
/*
                                  attach.c

Attachment file management.

soyMAIL manages attachments as files stored in the VMS Mail directory.  The
names of these files allow them to be distinguished from VMS Mail and other
other soyMAIL files.  They are named 'SOYMAIL_yyyymmddhhmmsshhnn.ATT' where
'yyyymmddhhmmsshh' are the numeric time components, and 'nn' a two-digit file
count.

The message compose page allows files stored on the user's browser system to be
uploaded to the system running soyMAIL.  These files are listed on the compose
page and can be marked (checkboxed) for inclusion as MIME parts in the sent
message. 

The message read page allows the MIME parts (usually attachments) of a received
message to be marked (checkboxed) and saved in the VMS Mail directory for
inclusion at the message composition page.

An attachment file has an internal structure.  The actual attachment content is
preceded by three fields with null-terminated strings explicitly providing the
soyMAIL version that saved them, the name and MIME content-type of the content.

  [soyMAIL-version]version\0
  [name]the name of the attachment.gif\0
  [type]image/gif\0
  [content]<bytes comprising the content until end-of-file>


COPYRIGHT
---------
Copyright (C) 2005,2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
25-MAR-2006  MGD  AttachView() small accomodation for Purveyor
15-MAR-2006  MGD  AttachFile() ensure any file system path is stripped
                    from the attachment name before saving (MSIE)
05-FEB-2006  MGD  support saving message as "message/rfc822" 'attachment'
                  bugfix; AttachComposeList() attachment size calculation
01-FEB-2005  MGD  initial
*/

/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unixio.h>

/* VMS related header files */
#include <descrip.h>
#include <rmsdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header file */
#include "soymail.h"
#include "attach.h"

#define FI_LI 	"ATTACH", __LINE__

/* global storage */

/* external storage */

extern BOOL  Debug,
             WatchEnabled;

extern char  SoftwareVn[];

extern VMS_MAIL_USER  VmsMailUser;

/* prototypes */
int sys$fao (__unknown_params);
int sys$numtim (__unknown_params);

/*****************************************************************************/
/*
Assemble the request data related to attachment management.  Return true to
indicate that there was attachment management undertaken and that the calling
function should do no more, false to indicate none and the calling function
should display a page.
*/

BOOL AttachComposeRequest (REQUEST_DATA *rdptr)

{
   int  cnt,
        AttachFileCount;
   char  *cptr, *sptr;
   char  AttachmentName [256],
         ButtonName [64];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AttachComposeRequest()\n");

   CGIVARNULL (cptr, "FORM_ATTACH_BTN");
   if (cptr)
   {
      /**********************/
      /* attachment buttons */
      /**********************/

      if (LangSame ("attach_file", cptr))
         AttachUpload (rdptr);
      else
      if (LangSame ("attach_delete", cptr))
         AttachDelete (rdptr, NULL);
      else
         ErrorExit (SS$_BUGCHECK, FI_LI);

      return (FALSE);
   }
   else
   {
      /**************************/
      /* individual attachments */
      /**************************/

      CGIVARNULL (cptr, "FORM_ATTACH_COUNT");
      if (cptr)
         AttachFileCount = atoi(cptr);
      else
         AttachFileCount = 0;

      for (cnt = 1; cnt <= AttachFileCount+5; cnt++)
      {
         sprintf (ButtonName, "FORM_ATTACH_DELETE_%d", cnt);
         CGIVARNULL (cptr, ButtonName);
         if (cptr)
         {
            sprintf (AttachmentName, "FORM_ATTACH_ID_%d", cnt);
            CGIVARNULL (sptr, AttachmentName);
            if (!sptr) ErrorExit (SS$_BUGCHECK, FI_LI);
            AttachDelete (rdptr, sptr);

            return (FALSE);
         }

         sprintf (ButtonName, "FORM_ATTACH_VIEW_%d", cnt);
         CGIVARNULL (cptr, ButtonName);
         if (cptr)
         {
            sprintf (AttachmentName, "FORM_ATTACH_ID_%d", cnt);
            CGIVARNULL (sptr, AttachmentName);
            if (!sptr) ErrorExit (SS$_BUGCHECK, FI_LI);

            /* NOTE: if successful this returns from here! */
            if (AttachView (rdptr, sptr)) return (TRUE);

            return (FALSE);
         }
      }
   }

   return (FALSE);
}

/*****************************************************************************/
/*
Generate a list of attachments available for sending from the message 
composition page. 
*/

void AttachComposeList (REQUEST_DATA *rdptr)

{
   int  cnt, status,
        AttachFileCount,
        AttachFileTotal,
        ButtonWidth,
        FileSizeBytes;
   char  *cptr, *fnptr, *sptr, *zptr;
   char  AttId [64],
         BlockBuffer [512],
         CbxName [256];
   stat_t  FstatBuffer;
   FILE  *fp;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AttachComposeList()\n");

   muptr = &VmsMailUser;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   cptr = LangFor("compose_button_width");
   ButtonWidth = atoi(cptr);
   if (ButtonWidth <= 0 || ButtonWidth > 10) ButtonWidth = 5;

   fprintf (stdout,
"<!-- message attachments begin -->\n\
<P CLASS=\"css_p08\">\
\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"5\" \
CLASS=\"css_panel_1\" ALIGN=\"center\">\n");

   /* first up count the number of attachment files */
   AttachFileTotal = 0;
   cptr = FileSpecSearch (1, muptr->VmsMailFullDirectory, "SOYMAIL_*.ATT");
   while (cptr)
   {
      AttachFileTotal++;
      cptr = FileSpecSearch (1, NULL, NULL);
   }

   /* list each one of those attachment files */
   AttachFileCount = 0;
   fnptr = FileSpecSearch (FALSE, muptr->VmsMailFullDirectory, "SOYMAIL_*.ATT");
   while (fnptr)
   {
      AttachFileCount++;

      /* get the digits from the file name */
      for (cptr = fnptr; *cptr; cptr++);
      while (cptr > fnptr && *cptr != ']') cptr--;
      if (*cptr == ']') cptr++;
      while (*cptr && !isdigit(*cptr)) cptr++;
      zptr = (sptr = AttId) + sizeof(AttId);
      while (isdigit(*cptr) && sptr < zptr) *sptr++ = *cptr++;
      if (sptr > zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      fp = fopen (fnptr, "r", "ctx=bin");
      if (!fp)
      {
         status = vaxc$errno;
         sprintf (BlockBuffer, "%s: %s", fnptr, SysGetMsg(status));
         cptr = BlockBuffer;
      }
      else
      {
         if (fstat (fileno(fp), &FstatBuffer) != 0) FstatBuffer.st_size = -1;
         FileSizeBytes = FstatBuffer.st_size;
         fread (BlockBuffer, sizeof(BlockBuffer), 1, fp);
         fclose (fp);
         if (!strncmp (BlockBuffer, "[soyMAIL-version]", 17))
         {
            sptr = BlockBuffer + 17;
            while (*sptr) sptr++;
            sptr++;
            if (!strncmp (sptr, "[name]", 6))
            {
               sptr += 6;
               cptr = sptr;
               /* adjust for the number of bytes in the internal data */
               while (*sptr) sptr++;
               sptr++;
               if (!strncmp (sptr, "[type]", 6))
               {
                  sptr += 6;
                  while (*sptr) sptr++;
                  sptr++;
                  if (!strncmp (sptr, "[content]", 9))
                  {
                     sptr += 9;
                     FileSizeBytes -= sptr - BlockBuffer;
                  }
                  else
                     cptr = "***ERROR4***";
               }
               else
                  cptr = "***ERROR3***";
            }
            else
               cptr = "***ERROR2***";
         }
         else
            cptr = "***ERROR1***";
      }

      /* a 'delete all' button on the first line of the attachment listing */
      if (AttachFileCount == 1)
         fprintf (stdout,
"<TR><TD WIDTH=\"10%%\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"attach_btn\" VALUE=\"%s\">\
</TD><TD COLSPAN=\"2\" STYLE=\"padding-left:15px;\">",
                  ButtonWidth,
                  LangFor("attach_delete"));
      else
         fprintf (stdout,
"<TR><TD WIDTH=\"10%%\"></TD><TD COLSPAN=\"2\" STYLE=\"padding-left:15px;\">");

      /* check if the attachment checkbox is currently checked */
      sprintf (CbxName, "FORM_PART_CBX_%s", AttId);
      CGIVARNULL (sptr, CbxName);
      if (sptr) sptr = " CHECKED";
      if (!sptr && !strcmp (rdptr->AttachmentName, cptr)) sptr = " CHECKED";
      if (!sptr) sptr = "";

      fprintf (stdout,
"<INPUT TYPE=\"hidden\" NAME=\"attach_id_%d\" VALUE=\"%s\">\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"attach_delete_%d\" VALUE=\"%s\">&nbsp;\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"attach_view_%d\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"attach_view_%d\" VALUE=\"^\" \
ONCLICK=\"targetBlank();\">\');\
</SCRIPT>\
&nbsp;&nbsp;&nbsp;\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"checkbox\" \
NAME=\"part_cbx_%s\" VALUE=\"1\"%s>\
&nbsp;\
%s\
&nbsp;&nbsp;<I>(%d %s)</I>\
</TD></TR>\n",
               AttachFileCount, AttId,
               AttachFileCount, LangFor("delete"),
               AttachFileCount, LangFor("view"),
               AttachFileCount, 
               AttId, 
               sptr,
               HTML_ESCAPE(cptr),
               FileSizeBytes > 1024 ? FileSizeBytes / 1024: FileSizeBytes,
               FileSizeBytes > 1024 ? "kB": "bytes");
      
      fnptr = FileSpecSearch (FALSE, NULL, NULL);
   }

   /* attachment file upload button */
   fprintf (stdout,
"<TR><TD WIDTH=\"10%%\" STYLE=\"vertical-align:top;\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"attach_btn\" VALUE=\"%s\" ONCLICK=\"startWorking();\">\
</TD><TD STYLE=\"padding-left:15px;\">\
<INPUT CLASS=\"css_text_1\" TYPE=\"file\" NAME=\"attach_file\" SIZE=\"50\">\
</TD><TD WIDTH=\"100%%\"></TD></TR>\n\
</TABLE>\n\
<INPUT TYPE=\"hidden\" NAME=\"attach_count\" VALUE=\"%d\">\n\
<!-- message attachments end -->\n\n",
               ButtonWidth,
               LangFor("attach_file"),
               AttachFileCount);
}                              

/*****************************************************************************/
/*
Save one or more attachment content (from the message read page).
*/

void AttachSave (REQUEST_DATA *rdptr)

{
   BOOL  AttachMessage = FALSE;
   int  cnt,
        DataLength,
        FileCount,
        PartCount;
   char  *cptr, *lptr, *sptr, *zptr,
         *DataPtr,
         *ListPtr,
         *NamePtr,
         *TypePtr;
   char  AttachCbxName [32];
   MIME_DATA  *mdptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AttachSave()\n");

   muptr = &VmsMailUser;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   uoptr = &rdptr->UserOptions;
   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;

   vmptr->MessageId = rdptr->MessageId;
   CallMailMessageGet (muptr, vmptr);
   if (VMSnok (muptr->MailVmsStatus))
   {
      StatusMessage (FI_LI, 1, "MAIL: %s.", SysGetMsg(muptr->MailVmsStatus));
      return;
   }

   CGIVARNULL (cptr, "FORM_PART_CBX_0");
   if (cptr) AttachMessage = TRUE;

   InetMailParseHeader (vmptr);
   if (vmptr->MimeDataPtr)
      if (!MimeDecParseBody (vmptr->MimeDataPtr))
         vmptr->MimeDataPtr = NULL;

   /* calculate maximum possible string space of the status message */
   cnt = 0;
   for (mdptr = vmptr->MimeDataPtr; mdptr; mdptr = mdptr->NextMimePtr)
   {
      if (mdptr->ContentTypeIsMultipart) continue;

      if (mdptr->NamePtr)
        cnt += strlen (mdptr->NamePtr);
      else
        cnt += strlen (mdptr->PartName);
      /* allow for ', ""' */
      cnt += 4;
   }
   if (AttachMessage) cnt += strlen(vmptr->Subject) + 4;

   ListPtr = lptr = calloc (cnt, 1);
   if (!ListPtr)  ErrorExit (vaxc$errno, FI_LI);

   FileCount = 0;

   if (AttachMessage)
   {
      /* the full message */
      if (AttachFile (vmptr->Subject, "message/rfc822",
                      vmptr->BodyPtr, vmptr->BodyLength))
      lptr += sprintf (lptr, "\"%s\"", vmptr->Subject);
      FileCount++;
   }

   CGIVAR (cptr, "FORM_PART_COUNT");
   if (cptr)
      PartCount = atoi(cptr);
   else
      PartCount = 0;

   for (mdptr = vmptr->MimeDataPtr; mdptr; mdptr = mdptr->NextMimePtr)
   {
      if (mdptr->ContentTypeIsMultipart) continue;

      for (cnt = 1; cnt <= PartCount; cnt++)
      {
         sprintf (AttachCbxName, "FORM_PART_CBX_%d", cnt);
         CGIVAR (cptr, AttachCbxName);
         if (!cptr) continue;
         if (mdptr->NamePtr)
            NamePtr = mdptr->NamePtr;
         else
            NamePtr = mdptr->PartName;
         if (!strsame (cptr, NamePtr, -1)) continue;

         TypePtr = mdptr->ContentTypePtr;

         if (MimeDecProcessPart (mdptr))
         {
            DataPtr = mdptr->PartContentPtr;
            DataLength = mdptr->PartContentLength;

            /* if writing the attachment file was not successful */
            if (!AttachFile (NamePtr, TypePtr, DataPtr, DataLength)) return;

            lptr += sprintf (lptr, "%s \"%s\"",
                             *ListPtr ? ", " : "", NamePtr);
            FileCount++;
         }
         break;
      }
   }

   if (FileCount)
      StatusMessage (FI_LI, 0, "%s %s.", LangFor("attach_save_done"),
                     HTML_ESCAPE(ListPtr));
   else
      StatusMessage (FI_LI, 1, "%s", LangFor("attach_none_supplied"));
}                              

/*****************************************************************************/
/*
Upload an attachment content (from the message composition page).
*/

void AttachUpload (REQUEST_DATA *rdptr)

{
   int  DataLength;
   char  *cptr, *sptr, *zptr,
         *DataPtr,
         *NamePtr,
         *TypePtr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AttachUpload()\n");

   muptr = &VmsMailUser;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   DataPtr = CgiLibVar ("FORM_ATTACH_FILE");
   if (WatchEnabled)
      WatchThis ("CGIVAR !AZ=!AZ", "FORM_ATTACH_FILE", DataPtr); 
   if (CgiLibBodyIsMultipartFormData())
      DataLength = CgiLibHtmlDeEntify (DataPtr);
   else
      DataLength = strlen(DataPtr);

   CGIVAR (NamePtr, "FORM_ATTACH_FILE_MIME_FILENAME");
   if (!(NamePtr && *NamePtr))
   {
      StatusMessage (FI_LI, 1, "%s", LangFor("attach_none_supplied"));
      return;
   }

   CGIVAR (TypePtr, "FORM_ATTACH_FILE_MIME_CONTENT_TYPE");
   if (!(TypePtr && *TypePtr)) TypePtr = "application/octet-stream";

   /* if the attachment file was not successfully writen */
   if (!AttachFile (NamePtr, TypePtr, DataPtr, DataLength)) return;

   /* save the file name for checkbox checking in the attachment list */
   zptr = (sptr = rdptr->AttachmentName) + sizeof(rdptr->AttachmentName)-1;
   for (cptr = NamePtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
   *sptr = '\0';

   StatusMessage (FI_LI, 0, "%s &quot;%s&quot;.",
                  LangFor("attach_upload_done"), NamePtr);
}                              

/*****************************************************************************/
/*
Write an attachment to the file system.
Return TRUE if the write was ok, FALSE otherwise.
*/

BOOL AttachFile
(
char *NamePtr,
char *TypePtr,
char *DataPtr,
int DataLength
)
{
   static $DESCRIPTOR (AttIdFaoDsc, "!4ZL!2ZL!2ZL!2ZL!2ZL!2ZL!2ZL!2ZL\0");
   static $DESCRIPTOR (FileNameFaoDsc, "SOYMAIL_!AZ.ATT\0");
   static int  FileCount;
   static unsigned short  NumTime [7];

   int  retval, status;
   char  *cptr, *fnptr;
   char  AttId [20],
         FileName [64];
   FILE  *fp;
   VMS_MAIL_USER  *muptr;
   $DESCRIPTOR (AttIdDsc, AttId);
   $DESCRIPTOR (FileNameDsc, FileName);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AttachFile()\n");

   muptr = &VmsMailUser;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   /* ensure any file system path is stripped from the file name (MSIE) */
   for (cptr = NamePtr; *cptr && *cptr != '/'; cptr++);
   if (*cptr)
   {
      while (*cptr) cptr++;
      while (cptr > NamePtr && *cptr != '/') cptr--;
      if (*cptr == '/') cptr++;
      NamePtr = cptr;
   }
   else
   {
      for (cptr = NamePtr; *cptr && *cptr != '\\'; cptr++);
      if (*cptr)
      {
         while (*cptr) cptr++;
         while (cptr > NamePtr && *cptr != '\\') cptr--;
         if (*cptr == '\\') cptr++;
         NamePtr = cptr;
      }
   }

   if (!NumTime[0])
   {
      /* only generate the one time-stamp per request */
      sys$numtim (&NumTime, 0);
   }

   /* two digits */
   if (FileCount++ > 99) ErrorExit (SS$_BUGCHECK, FI_LI);

   /* create an attachment ID based on the time and counter */
   sys$fao (&AttIdFaoDsc, NULL, &AttIdDsc,
            NumTime[0], NumTime[1], NumTime[2], NumTime[3],
            NumTime[4], NumTime[5], NumTime[6], FileCount);

   sys$fao (&FileNameFaoDsc, NULL, &FileNameDsc, AttId);

   fnptr = CallMailFileIn (muptr, FileName, NULL);
   if (WatchEnabled) WatchThis ("ATTACHMENT !AZ", fnptr); 

   fp = fopen (fnptr, "w", "ctx=bin", "rfm=udf", "rat=none");
   if (!fp)
   {
      status = vaxc$errno;
      StatusMessage (FI_LI, 1, "%s: %s.", NamePtr, SysGetMsg(status));
      return (FALSE);
   }

   retval = write (fileno(fp), "[soyMAIL-version]", 17);
   /* include the terminating null */
   if (retval > 0) retval = write (fileno(fp), SoftwareVn,
                                               strlen(SoftwareVn)+1);

   if (retval > 0) retval = write (fileno(fp), "[name]", 6);
   /* include the terminating null */
   if (retval > 0) retval = write (fileno(fp), NamePtr, strlen(NamePtr)+1);

   if (retval > 0) retval = write (fileno(fp), "[type]", 6);
   /* include the terminating null */
   if (retval > 0) retval = write (fileno(fp), TypePtr, strlen(TypePtr)+1);

   if (retval > 0) retval = write (fileno(fp), "[content]", 9);
   if (retval > 0) retval = write (fileno(fp), DataPtr, DataLength);
   fclose (fp);

   if (retval <= 0)
   {
      status = vaxc$errno;
      StatusMessage (FI_LI, 1, "%s: %s.", NamePtr, SysGetMsg(status));
      /* remove remnants of failed write */
      while (!remove (fnptr));
      return (FALSE);
   }

   /* purge the versions back to a maximum of 1 */
   strcat (fnptr, ";-1");
   while (!remove (fnptr));

   return (TRUE);
}                              

/*****************************************************************************/
/*
Delete attachment files, a single one or all.
If 'FileName' is NULL delete all attachment files.
*/

void AttachDelete
(
REQUEST_DATA *rdptr,
char *AttId
)
{
   int  status;
   char  *cptr, *fnptr, *sptr, *zptr;
   char  FileName [64];
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AttachDelete() %s\n", AttId);

   muptr = &VmsMailUser;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   if (AttId)
   {
      zptr = (sptr = FileName) + sizeof(FileName)-1;
      for (cptr = "SOYMAIL_"; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = AttId; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = ".ATT"; *cptr && sptr < zptr; *sptr++ = *cptr++);
      *sptr = '\0';

      fnptr = CallMailFileIn (muptr, FileName, NULL);

      if (WatchEnabled) WatchThis ("ATTACHMENT DELETE !AZ", fnptr); 

      if (remove (fnptr))
      {
         status = vaxc$errno;
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(status));
         return;
      }
      /* remove all versions */
      while (!remove (fnptr));

      StatusMessage (FI_LI, 0, "%s.", LangFor("attach_delete_done"));
   }
   else
   {
      cptr = FileSpecSearch (1, muptr->VmsMailFullDirectory, "SOYMAIL_*.ATT");
      while (cptr)
      {
         if (WatchEnabled) WatchThis ("ATTACHMENT DELETE !AZ", fnptr); 

         fnptr = CallMailFileIn (muptr, cptr, NULL);

         if (remove (fnptr))
         {
            status = vaxc$errno;
            StatusMessage (FI_LI, 1, "%s: %s.", fnptr, SysGetMsg(status));
            return;
         }
         /* remove all versions */
         while (!remove (fnptr));

         cptr = FileSpecSearch (1, NULL, NULL);
      }

      StatusMessage (FI_LI, 0, "%s.", LangFor("attach_delete_all"));
   }
}                              

/*****************************************************************************/
/*
The message composition page contains buttons for viewing attachment content.
*/

BOOL AttachView
(
REQUEST_DATA *rdptr,
char *AttId
)
{
   static $DESCRIPTOR (FileNameFaoDsc, "SOYMAIL_!AZ.ATT\0");

   int  status,
        ContentLength,
        DataLength,
        FileSizeBytes;
   char  *cptr, *fnptr, *czptr,
         *ContentTypePtr,
         *DataPtr;
   char  DataBuffer [4096],
         FileName [64];
   stat_t  FstatBuffer;
   FILE  *fp;
   VMS_MAIL_USER  *muptr;
   $DESCRIPTOR (FileNameDsc, FileName);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AttachView() %s\n", FileName);

   muptr = &VmsMailUser;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   sys$fao (&FileNameFaoDsc, NULL, &FileNameDsc, AttId);
   fnptr = CallMailFileIn (muptr, FileName, NULL);
   if (WatchEnabled) WatchThis ("ATTACHMENT !AZ", fnptr); 

   fp = fopen (fnptr, "r", "ctx=bin");
   if (!fp)
   {
      status = vaxc$errno;
      StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(status));
      return (FALSE);
   }
   else
   {
      if (fstat (fileno(fp), &FstatBuffer)) FstatBuffer.st_size = -1;
      FileSizeBytes = FstatBuffer.st_size;
      DataLength = read (fileno(fp), DataBuffer, sizeof(DataBuffer));
   }

   czptr = (cptr = DataBuffer) + DataLength;
   if (strncmp (cptr, "[soyMAIL-version]", 17))
   {
      StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(SS$_BUGCHECK));
      return (FALSE);
   }
   cptr += 17;
   while (*cptr && cptr < czptr) cptr++;
   if (cptr < czptr) cptr++;
   if (strncmp (cptr, "[name]", 6))
   {
      StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(SS$_BUGCHECK));
      return (FALSE);
   }
   cptr += 6;
   while (*cptr && cptr < czptr) cptr++;
   if (cptr < czptr) cptr++;
   if (strncmp (cptr, "[type]", 6))
   {
      StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(SS$_BUGCHECK));
      return (FALSE);
   }
   cptr += 6;
   /* now should be pointing at the content type */
   ContentTypePtr = cptr;
   /* skip over the type */
   while (*cptr && cptr < czptr) cptr++;
   if (cptr < czptr) cptr++;
   if (strncmp (cptr, "[content]", 9))
   {
      StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(SS$_BUGCHECK));
      return (FALSE);
   }
   /* now should be pointing at the content itself (finally) */
   cptr += 9;
   ContentLength = FileSizeBytes - (cptr - DataBuffer);
   /* what remains in the data buffer */
   DataLength -= cptr - DataBuffer;

   CgiLibResponseHeader (200, ContentTypePtr,
                         "Content-Length: %d\n", ContentLength);

   if (CgiLibEnvironmentIsPurveyor()) CgiLibEnvironmentBinaryOut ();

   while (DataLength)
   {
      fwrite (cptr, DataLength, 1, stdout);
      DataLength = read (fileno(fp), cptr = DataBuffer, sizeof(DataBuffer));
   }

   if (CgiLibEnvironmentIsPurveyor()) CgiLibEnvironmentRecordOut ();

   fclose (fp);

   return (TRUE);
}                              

/*****************************************************************************/

