/*****************************************************************************/
/*
                                  message.c

Displays a list of messages in a folder.
Displays the contents of a message (header, body, links to attachments).
Displays a 'print-view' (more suitable for printing) of a message.
Displays the raw message (header+body, or header alone).
Deletes specified messages.


COPYRIGHT
---------
Copyright (C) 2005,2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
03-APR-2005  MGD  significant rework to rationalise public and private
                  MessageMassage() massaged to be massaged in situ!
30-MAR-2006  MGD  MessagePage() precede each message part/attachment with
                  an anchor/link to allow for right-click browser save-as
25-MAR-2006  MGD  MessagePgpSigned() and MessagePgpCut() remove PGP signature
                  MessagePart() small accomodation for Purveyor
14-MAR-2006  MGD  bugfix; MessageBodyIdentify() when generating plain from
                    HTML-only content call MimeDecProcessPart() to decode
                  bugfix; MessageBodyIdentify() default a non-MIME message
                    message to 8-bit, plain-text
10-MAR-2006  MGD  check return from MimeDecParseBody() and process accordingly
                  remove 'noisy' message count from [message] range button
04-MAR-2006  MGD  MessagePage() add MessageBodyButtons() after longer messages
25-FEB-2006  MGD  refine MessageBodyIdentify() to manufacture plain-text
                    content when there is only HTML content
07-FEB-2006  MGD  optional folder buttons on left-side of message listing
05-FEB-2006  MGD  support saving message as "message/rfc822" 'attachment'
01-FEB-2005  MGD  initial
*/

/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <mailmsgdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <time.h>

/* application header file */
#include "soymail.h"
#include "attach.h"
#include "callmail.h"
#include "cgilib.h"
#include "compose.h"
#include "config.h"
#include "html.h"
#include "mainmenu.h"
#include "other.h"
#include "options.h"
#include "message.h"
#include "request.h"

#define FI_LI __FILE__, __LINE__

/* number of message lines before second set of message disposition buttons */
#define MESSAGE_BUTTONS_AFTER 32

/********************/
/* external storage */
/********************/

extern BOOL  Debug,
             WatchEnabled;

extern int  VmsVersion;

extern char  *CgiEnvironmentPtr;

extern char  CurrentVmsTimeString[],
             SoftwareEnv[],
             SoftwareId[],
             SoftwareVn[];

extern CONFIG_DATA  SoyMailConfig;
extern USER_OPTIONS  UserOptions;
extern REQUEST_DATA  RequestData;
extern VMS_MAIL_USER  VmsMailUser;

/****************************************************************************/
/*
Parse the buttons associated with the folder message list.
*/ 

void MessagePageFolderRequest (REQUEST_DATA *rdptr)

{
   int  cnt, hash;
   char  *cptr, *sptr, *zptr,
         *MsgRangeListPtr;
   char  MsgBtnString [32],
         MsgCbxString [32];
   VMS_MAIL_USER  *muptr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessagePageFolderRequest()");

   /* a bit of sanity checking */
   if (!rdptr->PrivateAccess) ErrorExit (SS$_BUGCHECK, FI_LI);

   muptr = &VmsMailUser;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;
   uoptr = &rdptr->UserOptions;

   if (uoptr->MassageRead) rdptr->MassageMessage = TRUE;

   /* the string containing comma-separated message id dot hashes */
   CGIVAR (MsgRangeListPtr, "FORM_MSG_RANGE_LIST");

   /*****************************/
   /* individual message button */
   /*****************************/

   CGIVARNULL (cptr, "FORM_MSG_BTN");
   if (cptr)
   {
      rdptr->MessageId = atol(cptr);
      rdptr->MessageHash = MessageRangeString (rdptr->MessageId,
                                               MsgRangeListPtr);
      MessagePage (rdptr);
      return;
   }

   /************************/
   /* popup message button */
   /************************/

   CGIVARNULL (cptr, "FORM_MSG_BTN_POPUP");
   if (cptr && (cnt = atol(cptr)))
   {
      rdptr->MessageId = cnt;
      rdptr->MessageHash = MessageRangeString (rdptr->MessageId,
                                               MsgRangeListPtr);
      MessagePage (rdptr);
      return;
   }

   /*****************/
   /* message range */
   /*****************/

   /* the range button or range 'onchange=' */
   CGIVARNULL (cptr, "FORM_RANGE_MSGS");
   if (!cptr) CGIVARNULL (cptr, "FORM_RANGE_LIST_CHANGE");
   if (cptr && *cptr) CGIVARNULL (cptr, "FORM_RANGE_LIST");
   /* also check that the range list is not empty */
   if (cptr && *cptr)
   {
      rdptr->RangeIdStart = atoi(cptr);
      while (*cptr && *cptr != '-') cptr++;
      if (*cptr)
      {
         cptr++;
         rdptr->RangeIdCount = atoi(cptr);
         rdptr->RangeIdCount -= rdptr->RangeIdStart - 1;
      }
      else
         rdptr->RangeIdCount = uoptr->MessagesPerPage;
      rdptr->RangeIdVector = +1;
      MessageFolderPage (rdptr);
      return;
   }

   /******************************/
   /* previous range of messages */
   /******************************/

   CGIVARNULL (cptr, "FORM_RANGE_PREV");
   if (cptr)
   {
      /* get the lowest id in the list and start one less than that */
      rdptr->RangeIdStart = MessageRangeString (-1, MsgRangeListPtr);
      if (rdptr->RangeIdStart <= 1)
      {
         rdptr->RangeIdStart = sdptr->RangeIdStart;
         StatusMessage (FI_LI, 1, LangFor("no_more_messages"));
      }
      else
         rdptr->RangeIdStart--;
      rdptr->RangeIdCount = uoptr->MessagesPerPage;
      rdptr->RangeIdVector = -1;
      MessageFolderPage (rdptr);
      return;
   }

   /**************************/
   /* next range of messages */
   /**************************/

   CGIVARNULL (cptr, "FORM_RANGE_NEXT");
   if (cptr)
   {
      /* the highest id is always the first one in the string */
      rdptr->RangeIdStart = atoi(MsgRangeListPtr) + 1;
      if (rdptr->RangeIdStart >= sdptr->FolderMessageCount)
      {
         /* just go back to the start of whatever we had before */
         rdptr->RangeIdStart = sdptr->RangeIdStart;
         StatusMessage (FI_LI, 1, LangFor("no_more_messages"));
      }
      rdptr->RangeIdCount = uoptr->MessagesPerPage;
      rdptr->RangeIdVector = +1;
      MessageFolderPage (rdptr);
      return;
   }

   /*****************************/
   /* copy, move, waste, delete */
   /*****************************/

   CGIVARNULL (cptr, "FORM_MSG_COPY");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_MOVE");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_WASTE");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_DELETE");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_EMPTY");
   if (cptr)
   {
      MessageIdList (rdptr);

      CGIVARNULL (cptr, "FORM_MSG_COPY");
      if (cptr)
         MessageCopy (rdptr, FALSE);
      else
      {
         CGIVARNULL (cptr, "FORM_MSG_MOVE");
         if (cptr)
            MessageCopy (rdptr, TRUE);
         else
         {
            CGIVARNULL (cptr, "FORM_MSG_WASTE");
            if (cptr)
            {
               rdptr->DestFolderName[0] = '_';
               strcpy (rdptr->DestFolderName+1, LangFor("folder_soy_waste"));
               rdptr->DestFolderNameLength = strlen(rdptr->DestFolderName);
               MessageCopy (rdptr, TRUE);
            }
            else
            {
               CGIVARNULL (cptr, "FORM_MSG_EMPTY");
               if (cptr)
               {
                  /* all messages in soyWASTE folder! */
                  if (strcmp (rdptr->FolderName+1, LangFor("folder_soy_waste")))
                     ErrorExit (SS$_BUGCHECK, FI_LI);
                  rdptr->RangeIdStart = rdptr->RangeIdCount = 0;
                  rdptr->MsgIdListPtr = "*";
                  rdptr->MsgIdListCount = 1;
                  MessageDelete (rdptr);
               }
               else
               {
                  CGIVARNULL (cptr, "FORM_MSG_DELETE");
                  if (cptr)
                     MessageDelete (rdptr);
                  else
                     ErrorExit (SS$_BUGCHECK, FI_LI);
               }
            }
         }
      }

      rdptr->RangeIdStart = sdptr->RangeIdStart;
      rdptr->RangeIdCount = uoptr->MessagesPerPage;

      MessageFolderPage (rdptr);

      /* make sure (any) deleted messages are purged from the WASTEBASKET */
      CallMailFileClose (muptr);

      return;
   }

   /*****************/
   /* folder button */
   /*****************/

   /* this button is also checked for in RequestPrivateDefaults() */
   CGIVARNULL (cptr, "FORM_FOLDER_BTN");
   if (cptr)
   {
      /* the list was peremtorially generated by RequestPrivateDefaults() */
      if (rdptr->MsgIdListCount)
      {
         /* message checkboxes result in message moves to buttoned folder */
         RequestParseDestFileFolder (rdptr, cptr);
         MessageCopy (rdptr, TRUE);
      }

      MessageFolderPage (rdptr);

      return;
   }

   /*********************/
   /* none of the above */
   /*********************/

   rdptr->RangeIdStart = sdptr->RangeIdStart;
   rdptr->RangeIdCount = uoptr->MessagesPerPage;

   MessageFolderPage (rdptr);
}

/****************************************************************************/
/*
When 'MessageId' is positive find that message number in the string and return
it's associated hash.  Return zero if the message is not found (error).  When
'MessageId' is zero step through each of the message numbers in the string,
returning zero at end of string.  When 'MessageId' is negative find the first
message id (the last one in the string) and return it's message number.  Return
zero if the string is empty (error).
*/ 

int MessageRangeString
(
int MessageId,
char *RangeStringPtr
)
{
   static char  *LastPtr = NULL;

   int  id, hash;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "MessageRangeString() %d |%s|\n",
               MessageId, RangeStringPtr);

   if (MessageId > 0)
   {
      /* return the hash of the specified message id from the string */
      cptr = RangeStringPtr;
      while (*cptr)
      {
         id = atol(cptr);
         if (!id) return (0);
         if (id == MessageId)
         {
            while (*cptr && *cptr != '.') cptr++;
            if (*cptr != '.') return (0);
            cptr++;
            hash = atol(cptr);
            return (hash);
         }
         while (*cptr && *cptr != ',') cptr++;
         if (*cptr) cptr++;
      }
      return (0);
   }

   if (MessageId == 0)
   {
      /* return each of the message IDs in the string in turn */
      if (!LastPtr) LastPtr = RangeStringPtr;
      cptr = LastPtr;
      id = atol(cptr);
      if (!id)
      {
         LastPtr = NULL;
         return (0);
      }
      while (*cptr && *cptr != ',') cptr++;
      if (*cptr) cptr++;
      LastPtr = cptr;
      return (id);
   }

   if (MessageId < 0)
   {
      /* return the number of the very last message id in the string */
      cptr = RangeStringPtr;
      id = 0;
      while (*cptr)
      {
         id = atol(cptr);
         if (!id) return (0);
         while (*cptr && *cptr != ',') cptr++;
         if (*cptr) cptr++;
      }
      return (id);
   }
}

/****************************************************************************/
/*
Parse the buttons from the message display page.
*/ 

void MessagePageMsgRequest (REQUEST_DATA *rdptr)

{
   static char  SingleMessageList [32];

   char  *cptr, *sptr, *zptr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessagePageMsgRequest()");

   /* a bit of sanity checking */
   if (!rdptr->PrivateAccess) ErrorExit (SS$_BUGCHECK, FI_LI);

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;

   if (uoptr->MassageRead) rdptr->MassageMessage = TRUE;

   /****************************/
   /* previous/at/next message */
   /****************************/

   CGIVARNULL (cptr, "FORM_MSG_PREV");
   if (cptr)
   {
      rdptr->MessageId = sdptr->MessageId;
      if (rdptr->MessageId > 1)
         rdptr->MessageId--;
      else
         StatusMessage (FI_LI, 1, LangFor("no_more_messages"));
      rdptr->MessageHash = 0;
      MessagePage (rdptr);
      return;
   }

   CGIVARNULL (cptr, "FORM_MSG_FOLDER_AT");
   if (cptr)
   {
      rdptr->RangeIdCount = uoptr->MessagesPerPage;
      for (rdptr->RangeIdStart = sdptr->FolderMessageCount;
           rdptr->RangeIdStart >= sdptr->MessageId;
           rdptr->RangeIdStart -= rdptr->RangeIdCount);
      rdptr->RangeIdStart++;
      if (rdptr->RangeIdStart < 1) rdptr->RangeIdStart = 1;
      MessageFolderPage (rdptr);
      return;
   }

   CGIVARNULL (cptr, "FORM_MSG_NEXT");
   if (cptr)
   {
      rdptr->MessageId = sdptr->MessageId;
      if (rdptr->MessageId < sdptr->FolderMessageCount)
         rdptr->MessageId++;
      else
         StatusMessage (FI_LI, 1, LangFor("no_more_messages"));
      rdptr->MessageHash = 0;
      MessagePage (rdptr);
      return;
   }

   /********************/
   /* specific message */
   /********************/

   CGIVAR (cptr, "FORM_MSG_ID_HASH");
   rdptr->MessageId = atol(cptr);
   while (*cptr && *cptr != '.') cptr++;
   if (*cptr) cptr++;
   if (!*cptr) ErrorExit (SS$_BUGCHECK, FI_LI);;
   rdptr->MessageHash = atol(cptr);

   /**********************************/
   /* copy/move/waste/delete message */
   /**********************************/

   CGIVARNULL (cptr, "FORM_MSG_COPY");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_MOVE");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_WASTE");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_DELETE");
   if (cptr)
   {
      sprintf (SingleMessageList, "%d.%d",
               rdptr->MessageId, rdptr->MessageHash);
      rdptr->MsgIdListPtr = SingleMessageList;
      rdptr->MsgIdListCount = 1;

      CGIVARNULL (cptr, "FORM_MSG_COPY");
      if (cptr)
         MessageCopy (rdptr, FALSE);
      else
      {
         CGIVARNULL (cptr, "FORM_MSG_MOVE");
         if (cptr)
            MessageCopy (rdptr, TRUE);
         else
         {
            CGIVARNULL (cptr, "FORM_MSG_WASTE");
            if (cptr)
            {
               rdptr->DestFolderName[0] = '_';
               strcpy (rdptr->DestFolderName+1, LangFor("folder_soy_waste"));
               rdptr->DestFolderNameLength = strlen(rdptr->DestFolderName);
               MessageCopy (rdptr, TRUE);
            }
            else
            {
               CGIVARNULL (cptr, "FORM_MSG_DELETE");
               if (cptr)
                  MessageDelete (rdptr);
               else
                  ErrorExit (SS$_BUGCHECK, FI_LI);
            }
         }
         if (sdptr->MessageId == sdptr->FolderMessageCount)
            rdptr->MessageId = sdptr->MessageId - 1; 
      }

      if (VMSnok (muptr->MailVmsStatus))
      {
         /* same behaviour and formula as folder-at */
         rdptr->RangeIdCount = uoptr->MessagesPerPage;
         for (rdptr->RangeIdStart = sdptr->FolderMessageCount;
              rdptr->RangeIdStart >= sdptr->MessageId;
              rdptr->RangeIdStart -= rdptr->RangeIdCount);
         rdptr->RangeIdStart++;
         if (rdptr->RangeIdStart < 1) rdptr->RangeIdStart = 1;
         MessageFolderPage (rdptr);
      }
      else
      {
         /* adjust this to reflect the successful move/copy/delete */
         if (sdptr->FolderMessageCount) sdptr->FolderMessageCount--;

         if (rdptr->MessageId)
         {
            /* still have message(s) in the folder */
            rdptr->MessageHash = 0;
            MessagePage (rdptr);
         }
         else
         {
            /* folder empty */
            rdptr->RangeIdStart = RANGE_MOST_RECENT;
            rdptr->RangeIdCount = uoptr->MessagesPerPage;
            rdptr->FolderMessageCount = 0;
            MessageFolderPage (rdptr);
         }
      }

      /* make sure (any) deleted messages are purged from the WASTEBASKET */
      CallMailFileClose (muptr);

      return;
   }

   /*****************/
   /* message parts */
   /*****************/

   CGIVARNULL (cptr, "FORM_PART_BTN");
   if (cptr)
   {
      zptr = (sptr = rdptr->AttachmentName) + sizeof(rdptr->AttachmentName)-1;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      /* if the attachment is successfully accessed then just finish up */
      if (MessagePart (rdptr)) return;
      MessagePage (rdptr);
      return;
   }

   CGIVARNULL (cptr, "FORM_PART_SAVE");
   if (cptr)
   {
      AttachSave (rdptr);
      MessagePage (rdptr);
      return;
   }

   /*************************/
   /* message reply buttons */
   /*************************/

   CGIVARNULL (cptr, "FORM_MSG_REPLY");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_REPLY_POPUP");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_REPLY_ALL");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_REPLY_ALL_POPUP");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_FORWARD");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_FORWARD_POPUP");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_EDIT_NEW");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_EDIT_NEW_POPUP");
   if (cptr)
   {
      ComposePageRequest (rdptr);
      return;
   }

   /************************/
   /* message read buttons */
   /************************/

   CGIVARNULL (cptr, "FORM_MSG_PRINT");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_PRINT_POPUP");
   if (cptr)
   {
      if (MessagePrintView (rdptr)) return;
      MessagePage (rdptr);
      return;
   }

   CGIVARNULL (cptr, "FORM_MSG_RAW");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_RAW_POPUP");
   if (cptr)
   {
      if (MessageRawView (rdptr)) return;
      MessagePage (rdptr);
      return;
   }

   CGIVARNULL (cptr, "FORM_MSG_HDR");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_HDR_POPUP");
   if (cptr)
   {
      if (MessageHdrView (rdptr)) return;
      MessagePage (rdptr);
      return;
   }

   /**************/
   /* HTML/plain */
   /**************/

   CGIVARNULL (cptr, "FORM_MSG_READ_PREF");
   if (cptr)
   {
      /* one is on, minus one is off, zero is not set! */
      if (LangSame ("message_read_html", cptr))
         rdptr->MessageReadPreference = OPTIONS_READ_PLAIN2_HTML1;
      if (LangSame ("message_read_plain", cptr))
         rdptr->MessageReadPreference = OPTIONS_READ_PLAIN1_HTML2;
      MessagePage (rdptr);
      return;
   }

   /****************/
   /* message wrap */
   /****************/

   CGIVARNULL (cptr, "FORM_MSG_WRAP");
   if (cptr)
   {
      /* one is on, minus one is off, zero is not set! */
      if (LangSame ("message_massage_on", cptr))
         rdptr->MassageMessage = TRUE;
      else
      if (LangSame ("message_massage_off", cptr))
         rdptr->MassageMessage = FALSE;
      /* use the read preference from the previous request */
      rdptr->MessageReadPreference = sdptr->MessageReadPreference;
      MessagePage (rdptr);
      return;
   }

   /******************/
   /* add as contact */
   /******************/

   CGIVARNULL (cptr, "FORM_MSG_ADD_CONTACT");
   if (cptr)
   {
      CGIVARNULL (cptr, "FORM_MSG_ADD_CONTACT_AS");
      if (cptr && *cptr)
      {
         CGIVARNULL (sptr, "FORM_MSG_ADD_CONTACT_FROM");
         if (sptr && *sptr)
            ContactsAppend (rdptr, cptr, sptr);
         else
            ErrorExit (SS$_BUGCHECK, FI_LI);
      }
      else
         StatusMessage (FI_LI, 1, LangFor("message_add_contact_name"));

      MessagePage (rdptr);
      return;
   }

   /* whatever we were doing before (disabled search can do this) */
   MessagePage (rdptr);
}

/****************************************************************************/
/*
Generate a string containing a comma-separated list of message ID numbers (in
textual form of course) representing the checkboxes against each message on the
message list page.  Return the number of checkboxes checked.
*/ 

int MessageIdList (REQUEST_DATA *rdptr)

{
   int  cnt, hash, id;
   char  *cptr, *sptr,
         *MsgRangeListPtr;
   char  MsgCbxString [32];
   REQUEST_DATA  *sdptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MessageIdList()\n");

   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;

   rdptr->MsgIdListPtr = NULL;
   rdptr->MsgIdListCount = 0;
   CGIVARNULL (cptr, "FORM_MSG_CBX_ALL");
   if (cptr && *cptr)
   {
      /* all messages checkbox! */
      rdptr->RangeIdStart = rdptr->RangeIdCount = 0;
      rdptr->MsgIdListPtr = "*";
      rdptr->MsgIdListCount = 1;
   }
   else
   {
      /* the string containing comma-separated message id dot hashes */
      CGIVAR (MsgRangeListPtr, "FORM_MSG_RANGE_LIST");

      /* count the number of messages in the range */
      for (cnt = 0; MessageRangeString (0, MsgRangeListPtr); cnt++);

      if (cnt)
      {
         /* message checkboxes */
         rdptr->MsgIdListPtr = sptr = calloc (cnt, 32);
         if (!rdptr->MsgIdListPtr) ErrorExit (vaxc$errno, FI_LI);
         while (id = MessageRangeString (0, MsgRangeListPtr))
         {
            sprintf (MsgCbxString, "FORM_MSG_CBX_%d", id);
            CGIVARNULL (cptr, MsgCbxString);
            if (cptr)
            {
               hash = MessageRangeString (id, MsgRangeListPtr);
               sptr += sprintf (sptr, "%s%d.%d",
                                rdptr->MsgIdListCount ? "," : "", id, hash);
               rdptr->MsgIdListCount++;
            }
         }
         *sptr = '\0';
         if (WatchEnabled)
            WatchThis ("CBX !UL!AZ!AZ",
                       rdptr->MsgIdListCount,
                       rdptr->MsgIdListCount ? " " : "", rdptr->MsgIdListPtr);
      }
   }

   if (Debug)
      fprintf (stdout, "%d %d %d %d|%s|\n",
               rdptr->MessageId, rdptr->RangeIdStart, rdptr->RangeIdCount,
               rdptr->MsgIdListCount, rdptr->MsgIdListPtr);

   return (rdptr->MsgIdListCount);
}

/*****************************************************************************/
/*
Display the basic private folder and message list page.
*/

void MessageFolderPage (REQUEST_DATA *rdptr)

{
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled)
      WatchThis ("MessageFolderPage() !AZ !AZ !SL !SL !SL",
                 rdptr->MailFileName, rdptr->FolderName,
                 rdptr->RangeIdStart, rdptr->RangeIdCount,
                 rdptr->RangeIdVector);

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;
   vmptr = &muptr->VmsMailMsg;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;

   if (rdptr->SearchData.InProgress) SearchCriteria (rdptr);

   /* determine if we're changing folders */
   if (!strcmp (rdptr->MailFileName, sdptr->MailFileName) &&
       !strcmp (rdptr->FolderName, sdptr->FolderName))
   {
      /* same folder, restore the message history from the state data */
      memcpy (&rdptr->MessageReadHistory,
              &sdptr->MessageReadHistory,
              sizeof(rdptr->MessageReadHistory));
      rdptr->MessageReadHistoryCount = sdptr->MessageReadHistoryCount;
      rdptr->FolderMessageCount = sdptr->FolderMessageCount;
   }
   else
      rdptr->RangeIdCount = 0;

   if (!rdptr->RangeIdCount)
   {
      rdptr->RangeIdStart = RANGE_MOST_RECENT;
      rdptr->RangeIdCount = uoptr->MessagesPerPage;
      rdptr->RangeIdVector = -1;
   }

   MessageListInit (rdptr);

   MainMenuPageBegin (rdptr, PAGE_FOLDER);
   MainMenuBar (rdptr);
   StatusInfoPanel (rdptr);
   MessageFolderList (rdptr);

   MainMenuPageEnd (rdptr);
}

/*****************************************************************************/
/*
Display the message header and body page.
*/

void MessagePage (REQUEST_DATA *rdptr)

{
   int  idx, status,
        LineCount;
   MIME_DATA  *mdptr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;
   VMS_MAIL_USER  UserData;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessagePage()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;
   vmptr = &muptr->VmsMailMsg;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;

   if (rdptr->SearchData.InProgress) SearchCriteria (rdptr);

   rdptr->FolderMessageCount = sdptr->FolderMessageCount;

   vmptr->MessageId = rdptr->MessageId;
   vmptr->ConfirmHash = rdptr->MessageHash;

   CallMailMessageGet (muptr, NULL);

   /* in case a message hash was not supplied use the one just generated */
   rdptr->MessageHash = vmptr->MessageHash;

   if (VMSnok (muptr->MailVmsStatus))
   {
      if (vmptr->ErrorOpenIn)
      {
         /* error opening external message file */
         StatusMessage (FI_LI, 1, "MAIL: %s %s",
                        SysGetMsg(MAIL$_OPENIN), vmptr->ExtId);
      }
      else
      {
         if (muptr->MailVmsStatus == SS$_BADCHECKSUM ||
             muptr->MailVmsStatus == MAIL$_NOMOREMSG)
            StatusMessage (FI_LI, 1, "%s", LangFor("message_checksum"));
         else
            StatusMessage (FI_LI, 1, "MAIL: %s.",
                           SysGetMsg(muptr->MailVmsStatus));
      }
      MessageFolderPage (rdptr);
      return;
   }

   /* restore the message history from the state data */
   memcpy (&rdptr->MessageReadHistory,
           &sdptr->MessageReadHistory,
           sizeof(rdptr->MessageReadHistory));
   rdptr->MessageReadHistoryCount = sdptr->MessageReadHistoryCount;

   /* shuffle everything down one */
   if (rdptr->MessageReadHistory[0] != rdptr->MessageId)
      for (idx = rdptr->MessageReadHistoryCount; idx > 0; idx--)
         rdptr->MessageReadHistory[idx] = rdptr->MessageReadHistory[idx-1]; 
   rdptr->MessageReadHistory[0] = rdptr->MessageId;
   if (rdptr->MessageReadHistoryCount < MESSAGE_READ_HISTORY_MAX)
      rdptr->MessageReadHistoryCount++;

   /********************/
   /* display the page */
   /********************/

   MessageContentIdentify (rdptr);

   MainMenuPageBegin (rdptr, PAGE_MESSAGE);

   MainMenuBar (rdptr);

   StatusInfoPanel (rdptr);

   MessageHeaderButtons (rdptr);

   MessageHeader (rdptr);

   MessageBodyButtons (rdptr);

   LineCount = MessageBody (rdptr);

   if (LineCount > MESSAGE_BUTTONS_AFTER) MessageBodyButtons (rdptr);

   MessageAttachments (rdptr);

   MainMenuPageEnd (rdptr);
}

/*****************************************************************************/
/*
Check the message for textual content based on user preferences.  Once decided
generate a description of that and the overall message  for the status
information panel.
*/

void MessageContentIdentify (REQUEST_DATA *rdptr)

{
   int  cnt, status,
        ReadPreference;
   char  *sptr;
   char  Description [256],
         Scratch [128];
   MIME_DATA  *mdptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageContentIdentify()");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;
   uoptr = &rdptr->UserOptions;

   InetMailParseHeader (vmptr);
   if (vmptr->MimeDataPtr)
      if (!MimeDecParseBody (vmptr->MimeDataPtr))
         vmptr->MimeDataPtr = NULL;

   /**********************/
   /* determine the text */
   /**********************/

   /* 'probe' the content to see if there is an alternative */
   mdptr = MessageBodyIdentify (vmptr, OPTIONS_READ_PLAIN0_HTML1);
   if (mdptr) rdptr->HtmlTextIsPresent = TRUE;
   mdptr = MessageBodyIdentify (vmptr, OPTIONS_READ_PLAIN1_HTML0);
   if (mdptr) rdptr->PlainTextIsPresent = TRUE;

   if (rdptr->MessageReadPreference)
      ReadPreference = rdptr->MessageReadPreference;
   else
      ReadPreference = uoptr->ReadPreference;

   mdptr = MessageBodyIdentify (vmptr, ReadPreference);
   if (mdptr)
   {
      MimeDecProcessPart (mdptr);
      rdptr->CharSetPtr = mdptr->CharsetPtr;
      if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_PLAIN)
         rdptr->ReadPlainText = TRUE;
      else
      if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
         rdptr->ReadHtmlText = TRUE;
      else
      {
         StatusMessage (FI_LI, 1, "%s", LangFor("mime_type_unknown"));
         return;
      }
   }
   else
   if (!vmptr->VmsForeign)
   {
      StatusMessage (FI_LI, 1, "%s", LangFor("mime_type_unknown"));
      return;
   }

   /************************/
   /* describe the content */
   /************************/

   if (rdptr->PrivateAccess || rdptr->PublicWildcard)
      sprintf (Scratch, "%s&nbsp;/&nbsp;%04d",
               HTML_ESCAPE(rdptr->FolderName), rdptr->MessageId);
   else
      sprintf (Scratch, "%s&nbsp;/&nbsp;%04d",
               HTML_ESCAPE(rdptr->PublicAccessString), rdptr->MessageId);

   sptr += sprintf (sptr = Description, LangFor("message_opened"), Scratch);

   if (vmptr->MimeDataPtr)
      sptr += sprintf (sptr, "&nbsp MIME.");
   else
   if (vmptr->LooksLikeRfc)
      sptr += sprintf (sptr, "&nbsp RFC.");
   else
   if (vmptr->VmsForeign)
      sptr += sprintf (sptr, "&nbsp VMS/FOREIGN.");
   else
      sptr += sprintf (sptr, "&nbsp VMS.");

   if (mdptr)
      if (MessagePgpSigned (mdptr->PartContentPtr))
         sptr += sprintf (sptr, "&nbsp PGP.");

   if (rdptr->CharSetPtr &&
       !strsame (rdptr->CharSetPtr, uoptr->CharSet, -1))
      sptr += sprintf (sptr, "&nbsp %s.", rdptr->CharSetPtr);

   cnt = 0;
   for (mdptr = vmptr->MimeDataPtr; mdptr; mdptr = mdptr->NextMimePtr)
      if (mdptr->NamePtr) cnt++;
   if (cnt)
      sptr += sprintf (sptr, "&nbsp; %d %s.",
                       cnt, LangFor("message_attachments"));

   cnt = 0;
   for (mdptr = vmptr->MimeDataPtr; mdptr; mdptr = mdptr->NextMimePtr)
      if (!mdptr->ContentTypeIsMultipart && mdptr->PartName[0]) cnt++;
   if (cnt > 1)
      sptr += sprintf (sptr, "&nbsp; %d %s.",
                       cnt, LangFor("message_parts"));

   if (rdptr->ReadPlainText &&
       vmptr->MimeDataPtr &&
       vmptr->MimeDataPtr->HtmlToPlainMimePtr)
      sptr += sprintf (sptr, "&nbsp %s.", LangFor("message_html_to_plain"));

   if (vmptr->BodyLength < 2048)
      sptr += sprintf (sptr, "&nbsp %d bytes.", vmptr->BodyLength);
   else
      sptr += sprintf (sptr, "&nbsp %d kbytes.", vmptr->BodyLength >> 10);

   if (vmptr->ExtIdLength)
      sptr += sprintf (sptr, "\n<!-- %s -->", vmptr->ExtId);

   StatusMessage (FI_LI, 0, "%s", Description);
}

/*****************************************************************************/
/*
Display the message header ("Subject:", "From:", "Date:", etc).
*/

void MessageHeader (REQUEST_DATA *rdptr)

{
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageHeader()");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;

   fprintf (stdout,
"<!-- header begin -->\n\
<P CLASS=\"css_p08\">\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"0\" CELLSPACING=\"10\" \
CLASS=\"css_panel_3\">\n\
<TR><TD ALIGN=\"left\">\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"2\" BORDER=\"0\">\n\
<INPUT TYPE=\"hidden\" NAME=\"msg_id_hash\" VALUE=\"%d.%d\">\n\
<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\" WIDTH=\"100%%\">%s</TD>\
<TD CLASS=\"css_msgnum\">%04d&nbsp;/&nbsp;%04d</TD></TR>\n\
<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\">%s</TD>",
            vmptr->MessageId, vmptr->MessageHash,
            LangFor("subject"),
            vmptr->MsgSubjPtr[0] ? HTML_TEXT(vmptr->MsgSubjPtr) :
                                   LangFor("(none)"),
            vmptr->MessageId, rdptr->FolderMessageCount,
            LangFor("from"),
            vmptr->MsgFromPtr[0] ? HTML_ADDR(vmptr->MsgFromPtr) :
                                   LangFor("(none)"));

   if (rdptr->PrivateAccess)
      fprintf (stdout,
"<TD CLASS=\"css_addas\" ROWSPAN=\"2\">\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_add_contact\" VALUE=\"%s:\">&nbsp;\
<INPUT CLASS=\"css_text_1\" TYPE=\"text\" SIZE=\"15\" \
NAME=\"msg_add_contact_as\" VALUE=\"%s\">\
<INPUT CLASS=\"css_text_1\" TYPE=\"hidden\" \
NAME=\"msg_add_contact_from\" VALUE=\"%s\">\
</TD>",
               LangFor("message_add_contact"),
               HTML_ESCAPE(ContactsNameAs(vmptr->MsgFromPtr)),
               HTML_ESCAPE(InetMailStripTransport(vmptr->MsgFromPtr)));

   fprintf (stdout,
"</TR>\n\
<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\" COLSPAN=\"2\">%s%s%s%s</TD></TR>\n\
<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\" COLSPAN=\"2\">%s</TD></TR>\n",
            LangFor("date"), HTML_TEXT(vmptr->MsgDatePtr),
            vmptr->LooksLikeRfc ? "&nbsp; [" : "",
            vmptr->LooksLikeRfc ? InetMailRfcVmsDate(vmptr->Date) : "",
            vmptr->LooksLikeRfc ? "]" : "",
            LangFor("to"),
            vmptr->MsgToPtr[0] ? HTML_ADDR(vmptr->MsgToPtr) :
                                 LangFor("(none)"));

   if (vmptr->MsgCcPtr[0])
      fprintf (stdout,
"<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\" COLSPAN=\"2\">%s</TD></TR>\n",
               LangFor("cc"), HTML_ADDR(vmptr->MsgCcPtr));

   fprintf (stdout,
"</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
<!-- header end -->\n\n");

}

/*****************************************************************************/
/*
Provide the [print], [raw], header] buttons, etc.
*/

void MessageHeaderButtons (REQUEST_DATA *rdptr)

{
   char  *cptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageHeaderButtons()");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;

   fprintf (stdout,
"<!-- header buttons begin -->\n\
<P CLASS=\"css_p08\">\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"5\" \
ALIGN=\"center\" CLASS=\"css_panel_1\">\n\
\
<TR><TD STYLE=\"text-align:left;\">\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_print\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_print_popup\" VALUE=\"^\" \
ONCLICK=\"targetBlank();\">\');\
</SCRIPT>\n\
\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_raw\" VALUE=\"%s\"%s>\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_raw_popup\" VALUE=\"^\"%s>\');\
</SCRIPT>\n\
\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_hdr\" VALUE=\"%s\"%s>\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_hdr_popup\" VALUE=\"^\"%s>\');\
</SCRIPT>\n\
\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_prev\" VALUE=\"%s\"%s>\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_folder_at\" VALUE=\"%s\"%s>\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_next\" VALUE=\"%s\"%s>\n\
</TD><TD STYLE=\"text-align:right;\">\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_copy\" VALUE=\"%s &gt;&gt;\"%s>\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_move\" VALUE=\"%s &gt;&gt;\"%s>\n\
<SELECT CLASS=\"css_select_1\" NAME=\"folder_list_dest\" \
ONCHANGE=\"document.soymail_form.folder_text_dest.value=\'\'\">\n",
            LangFor("print"),
            LangFor("raw"),
            vmptr->LooksLikeRfc ? "" : "ONCLICK=\"return notApplicable();\"",
            vmptr->LooksLikeRfc ? "ONCLICK=\"targetBlank();\"" :
                                  "ONCLICK=\"return notApplicable();\"",
            LangFor("header"),
            vmptr->LooksLikeRfc ? "" : "ONCLICK=\"return notApplicable();\"",
            vmptr->LooksLikeRfc ? "ONCLICK=\"targetBlank();\"" :
                                  "ONCLICK=\"return notApplicable();\"",
            LangFor("prev"),
            CLICK_NOMORE_STARTWORKING(vmptr->MessageId == 1),
            LangFor("folder_at"),
            CLICK_STARTWORKING,
            LangFor("next"),
            CLICK_NOMORE_STARTWORKING(vmptr->MessageId ==
                                      rdptr->FolderMessageCount),
            LangFor("copy"),
            CLICK_STARTWORKING,
            LangFor("move"),
            CLICK_STARTWORKING);

   FolderDestListEach (rdptr);

   CGIVAR (cptr, "FORM_FOLDER_TEXT_DEST");

   fprintf (stdout,
"</SELECT>\n\
<INPUT CLASS=\"css_text_1\" TYPE=\"text\" \
NAME=\"folder_text_dest\" SIZE=\"8\" VALUE=\"%s\" \
ONCLICK=\"this.select()\">\n\
</TD></TR>\n\
</TABLE>\n\
<!-- header buttons end -->\n\n",
            HTML_ESCAPE(cptr));

}

/*****************************************************************************/
/*
Display the message body.
*/

int MessageBody (REQUEST_DATA *rdptr)

{
   BOOL  ReadPreference;
   int  status,
        LineCount = 0;
   char  *cptr;
   char  PartUri [512];
   MIME_DATA  *mdptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageBody()");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;
   uoptr = &rdptr->UserOptions;

   fprintf (stdout,
"<!-- message begin -->\n\
<SCRIPT LANGUAGE=\"JavaScript\">\n\
function resizeHTML(obj) {\n\
   docHeight = iframeHTML.document.body.scrollHeight;\n\
   obj.style.height = docHeight + \'px\';\n\
   obj.style.overflow = \'hidden\';\n\
}\n\
</SCRIPT>\n\
<P CLASS=\"css_p08\">\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"0\" CELLSPACING=\"10\" \
CLASS=\"css_panel_3\">\n\
<TR><TD>");

   if (rdptr->MessageReadPreference)
      ReadPreference = rdptr->MessageReadPreference;
   else
      ReadPreference = rdptr->UserOptions.ReadPreference;

   mdptr = MessageBodyIdentify (vmptr, ReadPreference);
   if (mdptr)
   {
      /* this in case a 'reply', etc., button is used */
      rdptr->ComposeData.ContentTypeIs = mdptr->ContentTypeIs;

      if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_PLAIN)
      {
         if (MimeDecProcessPart (mdptr))
         {
            LineCount = MessageLineCount (mdptr->PartContentPtr);

            fputs ("<PRE CLASS=\"css_text_message\">", stdout);

            if (rdptr->MassageMessage) MessageMassage (mdptr->PartContentPtr);
            fputs (HTML_ANCHOR(mdptr->PartContentPtr), stdout);

            fputs ("</PRE>", stdout);
         }
      }
      else
      if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      {
         if (MimeDecProcessPart (mdptr))
         {
            /* bit hit-and-miss for HTML but what the hey! */
            LineCount = MessageLineCount (mdptr->PartContentPtr);

            if (rdptr->PrivateAccess)
               sprintf (PartUri, "%s/%s/%s/%d.%d/--%s",
                        rdptr->FormAction,
                        rdptr->MailFileName,
                        HTML_ESCAPE(rdptr->FolderName),
                        rdptr->MessageId, rdptr->MessageHash,
                        HTML_ESCAPE(mdptr->PartName));
            else
            if (rdptr->PublicWildcard)
               sprintf (PartUri, "%s%s/%d.%d/--%s",
                        rdptr->FormAction,
                        HTML_ESCAPE(rdptr->FolderName),
                        rdptr->MessageId, rdptr->MessageHash,
                        HTML_ESCAPE(mdptr->PartName));
            else
               sprintf (PartUri, "%s%d.%d/--%s",
                        rdptr->FormAction,
                        rdptr->MessageId, rdptr->MessageHash,
                        HTML_ESCAPE(mdptr->PartName));

            fprintf (stdout,
"<IFRAME NAME=\"iframeHTML\" ID=\"iframeHTML\" \
FRAMEBORDER=\"0\" WIDTH=\"100%%\" STYLE=\"height:%dem;\" \
SRC=\"%s\" ONLOAD=\"parent.resizeHTML(this)\">\
*** NO &lt;IFRAME&gt;s! ***</IFRAME>",
                     rdptr->ClientHasJavaScript ? 0 : 20,
                     PartUri);
         }
      }
      else
         ErrorExit (SS$_BUGCHECK, FI_LI);
   }

   fprintf (stdout,
"</TD></TR>\n\
</TABLE>\n\
<!-- message end -->\n\n");

   return (LineCount);
}

/*****************************************************************************/
/*
Display (any) message attachment buttons.
*/

void MessageAttachments (REQUEST_DATA *rdptr)

{
   int  status,
        PartCount;
   char  PartUri [512];
   MIME_DATA  *mdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageAttachments()");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;

   if (vmptr->LooksLikeRfc)
      mdptr = vmptr->MimeDataPtr;
   else
      mdptr = NULL;

   if (mdptr)
   {
      /**************/
      /* MIME parts */
      /**************/

      if (rdptr->PrivateAccess)
         sprintf (PartUri, "%s/%s/%s/%d.%d/__%s",
                  rdptr->FormAction,
                  HTML_ESCAPE(rdptr->MailFileName),
                  HTML_ESCAPE(rdptr->FolderName),
                  rdptr->MessageId,
                  rdptr->MessageHash,
                  HTML_ESCAPE(LangFor("attach_message")));
      else
      if (rdptr->PublicWildcard)
         sprintf (PartUri, "%s%s/%d.%d/__%s",
                  rdptr->FormAction,
                  HTML_ESCAPE(rdptr->FolderName),
                  rdptr->MessageId, rdptr->MessageHash,
                  HTML_ESCAPE(LangFor("attach_message")));
      else
         sprintf (PartUri, "%s%d.%d/__%s",
                  rdptr->FormAction,
                  rdptr->MessageId, rdptr->MessageHash,
                  HTML_ESCAPE(LangFor("attach_message")));

      fprintf (stdout,
"<!-- attachment buttons begin -->\n\
<P CLASS=\"css_p08\">\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"5\" \
CLASS=\"css_panel_1\">\n\
<TR><TD STYLE=\"text-align:right;\" WIDTH=\"40%%\">\
<A CLASS=\"css_msg_part_link\" HREF=\"%s\" \
ONCLICK=\"return false\">%s</A>&nbsp;\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"part_btn\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
var part0URL=\"%s?\";\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"part_btn\" VALUE=\"^\" \
ONCLICK=\"targetBlank();\
window.open(part0URL+document.soymail_form.part_type_0.value);\
return false;\">\');\
</SCRIPT>&nbsp;\
</TD><TD STYLE=\"text-align:left;\" WIDTH=\"60%%\">\
<INPUT CLASS=\"css_text_1\" type=\"text\" SIZE=\"24\" \
NAME=\"part_type_0\" VALUE=\"message/rfc822\" \
ONFOCUS=\"this.select()\">\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"checkbox\" \
NAME=\"part_cbx_0\" VALUE=\"0\">\
&nbsp;&nbsp;",
               PartUri,
               MESSAGE_PART_LINK,
               LangFor("attach_message"),
               PartUri);

      if (rdptr->PrivateAccess)
         fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"part_save\" VALUE=\"%s\">&nbsp;",
                  LangFor ("attach_save_part"));

      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"reset\" VALUE=\"%s\">\
</TD></TR>\n",
               LangFor ("reset"));

      /* if it's just a plain-text message then don't bother */
      if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_PLAIN) mdptr = NULL;

      PartCount = 0;
      while (mdptr)
      {
         if (!mdptr->ContentTypeIsMultipart)
         {
            PartCount++;

            if (rdptr->PrivateAccess)
               sprintf (PartUri, "%s/%s/%s/%d.%d/%s",
                        rdptr->FormAction,
                        HTML_ESCAPE(rdptr->MailFileName),
                        HTML_ESCAPE(rdptr->FolderName),
                        rdptr->MessageId,
                        rdptr->MessageHash,
                        mdptr->NamePtr ? HTML_ESCAPE(mdptr->NamePtr) :
                                         HTML_ESCAPE(mdptr->PartName));
            else
            if (rdptr->PublicWildcard)
               sprintf (PartUri, "%s%s/%d.%d/%s",
                        rdptr->FormAction,
                        HTML_ESCAPE(rdptr->FolderName),
                        rdptr->MessageId, rdptr->MessageHash,
                        mdptr->NamePtr ? HTML_ESCAPE(mdptr->NamePtr) :
                                         HTML_ESCAPE(mdptr->PartName));
            else
               sprintf (PartUri, "%s%d.%d/%s",
                        rdptr->FormAction,
                        rdptr->MessageId, rdptr->MessageHash,
                        mdptr->NamePtr ? HTML_ESCAPE(mdptr->NamePtr) :
                                         HTML_ESCAPE(mdptr->PartName));

            fprintf (stdout,
"<TR><TD STYLE=\"text-align:right;\" WIDTH=\"40%%\">\
<A CLASS=\"css_msg_part_link\" HREF=\"%s\" \
ONCLICK=\"return false\">%s</A>&nbsp;\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"part_btn\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
var part%dURL=\"%s?\";\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"part_btn\" VALUE=\"^\" \
ONCLICK=\"targetBlank();\
window.open(part%dURL+document.soymail_form.part_type_%d.value);\
return false;\">\');\
</SCRIPT>&nbsp;\
</TD><TD STYLE=\"text-align:left;\" WIDTH=\"60%%\">\
<INPUT CLASS=\"css_text_1\" type=\"text\" SIZE=\"24\" \
NAME=\"part_type_%d\" VALUE=\"%s\" \
ONFOCUS=\"this.select()\">\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"checkbox\" \
NAME=\"part_cbx_%d\" VALUE=\"%s\">\
</TD></TR>\n",
                  PartUri,
                  MESSAGE_PART_LINK,
                  mdptr->NamePtr ? HTML_ESCAPE(mdptr->NamePtr) :
                                   HTML_ESCAPE(mdptr->PartName),
                  PartCount,
                  PartUri,
                  PartCount,
                  PartCount,
                  PartCount,
                  mdptr->ContentTypePtr ? HTML_ESCAPE(mdptr->ContentTypePtr) :
                                          "application/octet-stream",
                  PartCount,
                  mdptr->NamePtr ? HTML_ESCAPE(mdptr->NamePtr) :
                                   HTML_ESCAPE(mdptr->PartName));
         }
         mdptr = mdptr->NextMimePtr;
      }

      fprintf (stdout,
"</TABLE>\n\
<INPUT TYPE=\"hidden\" NAME=\"part_count\" VALUE=\"%d\">\n\
<!-- attachment buttons end -->\n\n",
               PartCount);
   }

   /***************/
   /* VMS/FOREIGN */
   /***************/

   if (vmptr->VmsForeign)
   {
      if (rdptr->PrivateAccess)
         sprintf (PartUri, "%s/%s/%s/%d.%d/FOREIGN",
                  rdptr->FormAction,
                  HTML_ESCAPE(rdptr->MailFileName),
                  HTML_ESCAPE(rdptr->FolderName),
                  rdptr->MessageId,
                  rdptr->MessageHash);
      else
      if (rdptr->PublicWildcard)
         sprintf (PartUri, "%s%s/%d.%d/FOREIGN",
                  rdptr->FormAction,
                  HTML_ESCAPE(rdptr->FolderName),
                  rdptr->MessageId, rdptr->MessageHash);
      else
         sprintf (PartUri, "%s%d.%d/FOREIGN",
                  rdptr->FormAction,
                  rdptr->MessageId, rdptr->MessageHash);

      fprintf (stdout,
"<!-- foreign button begin -->\n\
<TR><TD NOWRAP ALIGN=\"left\" VALIGN=\"middle\">\n\
<P CLASS=\"css_p08\">\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"5\" \
CLASS=\"css_panel_1\">\n\
<TR><TD STYLE=\"text-align:left;\">\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"part_btn\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
var partURL = \"%s\";\n\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"part_btn\" VALUE=\"^\" \
ONCLICK=\"targetBlank();window.open(partURL);return false;\">\');\
</SCRIPT>&nbsp;\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
<!-- foreign button end -->\n\n",
               LangFor("vms_foreign"),
               PartUri);
   }
}

/*****************************************************************************/
/*
Display the [reply], [reply to all], [forward], etc., buttons.
*/

void MessageBodyButtons (REQUEST_DATA *rdptr)

{
   int  status;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageBodyButtons()");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;

   fprintf (stdout,
"<!-- message buttons begin -->\n\
<P CLASS=\"css_p08\">\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"5\" \
CLASS=\"css_panel_1\">\n\
<TR><TD STYLE=\"text-align:left;\">\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" NAME=\"msg_reply\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_reply_popup\" VALUE=\"^\" \
ONCLICK=\"targetBlank();\">\');\
</SCRIPT>\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" NAME=\"msg_reply_all\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_reply_all_popup\" VALUE=\"^\" \
ONCLICK=\"targetBlank();\">\');\
</SCRIPT>\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" NAME=\"msg_forward\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_forward_popup\" VALUE=\"^\" \
ONCLICK=\"targetBlank();\">\');\
</SCRIPT>\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" NAME=\"msg_edit_new\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_edit_new_popup\" VALUE=\"^\" \
ONCLICK=\"targetBlank();\">\');\
</SCRIPT>\n",
            LangFor("reply"),
            LangFor("reply_all"),
            LangFor("forward"),
            LangFor("edit_as_new"));

   if (rdptr->ReadPlainText && rdptr->HtmlTextIsPresent)
   {
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_read_pref\" VALUE=\"%s\">\n",
               LangFor("message_read_html"));
   }
   else
   if (rdptr->ReadHtmlText && rdptr->PlainTextIsPresent)
   {
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\"  \
NAME=\"msg_read_pref\" VALUE=\"%s\">\n",
               LangFor("message_read_plain"));
   }

   if (rdptr->ReadPlainText)
   {
      /* one is on, minus one is off, zero is not set! */
      if (rdptr->MassageMessage)
         fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_wrap\" VALUE=\"%s\">\n", LangFor("message_massage_off"));
      else
         fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_wrap\" VALUE=\"%s\">\n", LangFor("message_massage_on"));
   }

   fprintf (stdout,
"</TD><TD STYLE=\"text-align:right;\">\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_delete\" VALUE=\"%s\"%s>\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_waste\" VALUE=\"%s\"%s>\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_prev\" VALUE=\"%s\"%s>\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_folder_at\" VALUE=\"%s\"%s>\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_next\" VALUE=\"%s\"%s>\n",
            LangFor("delete"),
            CLICK_STARTWORKING,
            LangFor("waste"),
            CLICK_STARTWORKING,
            LangFor("prev"),
            CLICK_NOMORE_STARTWORKING(vmptr->MessageId == 1),
            LangFor("folder_at"),
            CLICK_STARTWORKING,
            LangFor("next"),
            CLICK_NOMORE_STARTWORKING(vmptr->MessageId ==
                                      rdptr->FolderMessageCount));

   fprintf (stdout,
"</TD></TR>\n</TABLE>\n\
<!-- message buttons end -->\n\n");
}

/*****************************************************************************/
/*
Provide a message in a format more suitable for printing.
*/

BOOL MessagePrintView (REQUEST_DATA *rdptr)

{
   int  cnt, status,
        ReadPreference;
   char  *cptr;
   char  MessageDescription [512];
   MIME_DATA  *mdptr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;
   VMS_MAIL_USER  UserData;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessagePrintView()");

   uoptr = &rdptr->UserOptions;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;
   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;

   /* use the read preference from the previous request */
   if (sdptr->MessageReadPreference)
      ReadPreference = sdptr->MessageReadPreference;
   else
      ReadPreference = uoptr->ReadPreference;

   vmptr->MessageId = rdptr->MessageId;
   vmptr->ConfirmHash = rdptr->MessageHash;

   CallMailMessageGet (muptr, NULL);
   if (VMSnok (muptr->MailVmsStatus))
   {
      if (muptr->MailVmsStatus == SS$_BADCHECKSUM ||
          muptr->MailVmsStatus == MAIL$_NOMOREMSG)
         StatusMessage (FI_LI, 1, "%s", LangFor("message_checksum"));
      else
         StatusMessage (FI_LI, 1, "MAIL: %s.", SysGetMsg(muptr->MailVmsStatus));
      return (FALSE);
   }

   InetMailParseHeader (vmptr);
   if (vmptr->MimeDataPtr)
      if (!MimeDecParseBody (vmptr->MimeDataPtr))
         vmptr->MimeDataPtr = NULL;

   if (rdptr->PrivateAccess)
      sprintf (MessageDescription, "%s / %s / %s / %04d",
               rdptr->UserName, rdptr->MailFileName,
               rdptr->FolderName, rdptr->MessageId);
   else
   if (rdptr->PublicAccess)
      sprintf (MessageDescription, "%s / %04d",
               rdptr->PublicAccessString, rdptr->MessageId);
   else
      MessageDescription[0] = '\0';

   mdptr = MessageBodyIdentify (vmptr, ReadPreference);
   if (mdptr)
      rdptr->CharSetPtr = mdptr->CharsetPtr;
   else
   if (!vmptr->VmsForeign)
      StatusMessage (FI_LI, 1, "%s", LangFor("mime_type_unknown"));

   if (rdptr->CharSetPtr) CgiLibResponseSetCharset (rdptr->CharSetPtr);

   CgiLibResponseHeader (200, "text/html");

   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" VALUE=\"%s\">\n\
<META NAME=\"environment\" VALUE=\"%s\">\n",
            SoftwareEnv,
            CgiEnvironmentPtr);

   SoyMailStyleStuff (rdptr, "print");

   /* setTimeout() gives Firefox 1.5 a chance to render the page */
   fprintf (stdout,
"<TITLE>%s</TITLE>\n\
<SCRIPT LANGUAGE=\"JavaScript\">\n\
function resizeHTML(obj) {\n\
   docHeight = iframeHTML.document.body.scrollHeight;\n\
   obj.style.height = docHeight + \'px\';\n\
}\n\
</SCRIPT>\n\
</HEAD>\n\n\
<BODY ONLOAD=\"setTimeout(\'print()\',500);\">\n\
\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"5\">\n\n",
            SoyMailPageTitle(rdptr));

   if (SoyMailConfig.PrintHeader)
   {
      /***************/
      /* page header */
      /***************/

      /* if it appears to contain it's own markup then use that */
      if (strchr (SoyMailConfig.PrintHeader, '<'))
         cptr = "";
      else
         cptr =
" STYLE=\"font-size:120%;font-weight:bold;\
text-decoration:underline;text-align:center;\"";

      fprintf (stdout,
"<!-- page header begin -->\n\
<TR><TD COLSPAN=\"4\">\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\" WIDTH=\"100%%\">\n\
<TR><TD%s>%s</TD></TR>\n\
</TABLE><P CLASS=\"css_p08\">\n\
</TD></TR>\n\
<!-- page header end -->\n\n",
               cptr, SoyMailConfig.PrintHeader);
   }

   /******************/
   /* message header */
   /******************/

   fprintf (stdout,
"<!-- message begin -->\n\
<TR><TD>\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"0\" CELLSPACING=\"2\" BORDER=\"0\">\n\
<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\" WIDTH=\"100%%\" COLSPAN=\"2\">%s</TD>\
<TD CLASS=\"css_rfcval\" \
STYLE=\"white-space:nowrap;text-align:right;\">%s</TD></TR>\n\
<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\" COLSPAN=3>%s</TD></TR>\n\
<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\" COLSPAN=2 STYLE=\"white-space:nowrap;\">%s%s%s%s</TD>\
<TD WIDTH=\"100%%\"></TD></TR>\n\
<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\" COLSPAN=3>%s</TD></TR>\n",
            LangFor("subject"),
            vmptr->MsgSubjPtr[0] ? HTML_TEXT(vmptr->MsgSubjPtr) :
                                   LangFor("(none)"),
            InetMailRfcVmsDate(CurrentVmsTimeString),
            LangFor("from"),
            vmptr->MsgFromPtr[0] ? HTML_ADDR(vmptr->MsgFromPtr) :
                                   LangFor("(none)"),
            LangFor("date"), HTML_TEXT(vmptr->MsgDatePtr),
            vmptr->LooksLikeRfc ? "&nbsp; [" : "",
            vmptr->LooksLikeRfc ? InetMailRfcVmsDate(vmptr->Date) : "",
            vmptr->LooksLikeRfc ? "]" : "",
            LangFor("to"),
            vmptr->MsgToPtr[0] ? HTML_ADDR(vmptr->MsgToPtr) :
                                 LangFor("(none)"));

   if (vmptr->MsgCcPtr[0])
      fprintf (stdout,
"<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\" COLSPAN=3>%s</TD></TR>\n",
               LangFor("cc"), HTML_ADDR(vmptr->MsgCcPtr));

   fprintf (stdout,
"<TR><TD CLASS=\"css_rfchdr\">%s:</TD>\
<TD CLASS=\"css_rfcval\" COLSPAN=3>%s</TD></TR>\n\
<TR><TD COLSPAN=\"4\" WIDTH=\"100%%\"><HR ALIGN=\"left\" \
WIDTH=\"90%%\" SIZE=\"1\" NOSHADE CLASS=\"css_print_hr\"></TD></TR>\n\
\n\
<TR><TD COLSPAN=\"4\"><DIV ID=\"message_text_id\">",
            LangFor("message"), HTML_ESCAPE(MessageDescription));

   mdptr = MessageBodyIdentify (vmptr, ReadPreference);
   if (mdptr)
   {
      if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_PLAIN)
      {
         if (MimeDecProcessPart (mdptr))
         {
            fputs ("<PRE CLASS=\"css_text_message\" \
STYLE=\"overflow:visible;\">", stdout);

            if (sdptr->MassageMessage) MessageMassage (mdptr->PartContentPtr);
            fputs (HTML_ANCHOR(mdptr->PartContentPtr), stdout);

            fputs ("</PRE>", stdout);
         }
      }
      else
      if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      {
         if (rdptr->ClientHasJavaScript)
         {
            fprintf (stdout,
"\n<IFRAME NAME=\"iframeHTML\" ID=\"iframeHTML\" \
FRAMEBORDER=\"0\" WIDTH=\"100%%\" STYLE=\"height:0em;\"");

            if (rdptr->PrivateAccess)
               fprintf (stdout, " SRC=\"%s/%s/%s/%d.%d/--%s\" ",
                        rdptr->FormAction,
                        rdptr->MailFileName, rdptr->FolderName,
                        rdptr->MessageId, rdptr->MessageHash,
                        HTML_ESCAPE(mdptr->PartName));
            else
            if (rdptr->PublicAccess)
               fprintf (stdout, " SRC=\"%s%d.%d/--%s\" ",
                        rdptr->FormAction,
                        rdptr->MessageId, rdptr->MessageHash,
                        HTML_ESCAPE(mdptr->PartName));

            fprintf (stdout,
" ONLOAD=\"parent.resizeHTML(this)\">*** NO &lt;IFRAME&gt;s! ***</IFRAME>\n");
         }
         else
         {
            if (MimeDecProcessPart (mdptr))
            {
               if (uoptr->MassageRead)
                  mdptr->PartContentLength =
                     MessageRemovePgpSig (mdptr->PartContentPtr);

               HtmlSanitise (mdptr->PartContentPtr);
               fputs (mdptr->PartContentPtr, stdout);
            }
         }
      }
      else
         ErrorExit (SS$_BUGCHECK, FI_LI);
   }

   fprintf (stdout,
"</DIV></TD></TR>\n\
<TR><TD COLSPAN=\"4\" WIDTH=\"100%%\"><HR ALIGN=\"left\" \
WIDTH=\"90%%\" SIZE=\"1\" NOSHADE CLASS=\"css_print_hr\"></TD></TR>\n\
<!-- message end -->\n\n");

   /*******************/
   /* any attachments */
   /*******************/

   for (mdptr = vmptr->MimeDataPtr; mdptr; mdptr = mdptr->NextMimePtr)
      if (mdptr->NamePtr || mdptr->FileNamePtr) break;

   if (mdptr)
   {
      fprintf (stdout,
"<!-- attachments begin -->\n\
<TR><TD COLSPAN=\"4\">\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"2\" BORDER=\"0\" WIDTH=\"100%%\">\n");

      while (mdptr)
      {
         fprintf (stdout,
"<TR><TD CLASS=\"css_tabhdr\" \
STYLE=\"padding-left:20;white-space:nowrap;\">&lt;%s&gt;</TD>\
<TD CLASS=\"css_tabdat\" \
STYLE=\"padding-left:5;font-style:italic;\">%s</TD></TR>\n",
                  mdptr->NamePtr ? mdptr->NamePtr : mdptr->FileNamePtr,
                  mdptr->ContentTypePtr ? mdptr->ContentTypePtr : "?");

         mdptr = mdptr->NextMimePtr;
      }

      fprintf (stdout,
"</TABLE>\n\
</TD></TR>\n\
<!-- attachments end -->\n\n");
   }

   if (SoyMailConfig.PrintFooter)
   {
      /***************/
      /* page footer */
      /***************/

      /* if it appears to contain it's own markup then use that */
      if (strchr (SoyMailConfig.PrintFooter, '<'))
         cptr = "";
      else
         cptr = " STYLE=\"font-weight:bold;text-align:center;\"";

      fprintf (stdout,
"<!-- page footer begin -->\n\
<TR><TD COLSPAN=\"4\"><P CLASS=\"css_p08\">\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\" WIDTH=\"100%%\">\n\
<TR><TD%s>%s</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
<!-- page footer end -->\n\n",
               cptr, SoyMailConfig.PrintFooter);
   }

   /************/
   /* page end */
   /************/

   fprintf (stdout,
"</TABLE>\n\
</BODY>\n\
</HTML>\n");

   return (TRUE);
}

/*****************************************************************************/
/*
Display the entire mail message (header and body) as a plain text document.
*/

BOOL MessageRawView (REQUEST_DATA *rdptr)

{
   int  status;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageRawView()");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;

   vmptr->MessageId = rdptr->MessageId;
   vmptr->ConfirmHash = rdptr->MessageHash;

   CallMailMessageGet (muptr, NULL);
   if (VMSnok (muptr->MailVmsStatus))
   {
      if (muptr->MailVmsStatus == SS$_BADCHECKSUM ||
          muptr->MailVmsStatus == MAIL$_NOMOREMSG)
         StatusMessage (FI_LI, 1, "%s", LangFor("message_checksum"));
      else
         StatusMessage (FI_LI, 1, "MAIL: %s.", SysGetMsg(muptr->MailVmsStatus));
      return (FALSE);
   }

   InetMailParseHeader (vmptr);

   if (!vmptr->LooksLikeRfc)
   {
      StatusMessage (FI_LI, 1, "%s", LangFor("not_applicable"));
      return (FALSE);
   }

   CgiLibResponseHeader (200, "text/plain");

   fputs (vmptr->BodyPtr, stdout);

   return (TRUE);
}

/*****************************************************************************/
/*
Display the RFC822/2822 header as a plain text document.
*/

BOOL MessageHdrView (REQUEST_DATA *rdptr)

{
   int  status;
   char  ch;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageHdrView()");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;

   vmptr->MessageId = rdptr->MessageId;
   vmptr->ConfirmHash = rdptr->MessageHash;

   CallMailMessageGet (muptr, NULL);
   if (VMSnok (muptr->MailVmsStatus))
   {
      if (muptr->MailVmsStatus == SS$_BADCHECKSUM ||
          muptr->MailVmsStatus == MAIL$_NOMOREMSG)
         StatusMessage (FI_LI, 1, "%s", LangFor("message_checksum"));
      else
         StatusMessage (FI_LI, 1, "MAIL: %s.", SysGetMsg(muptr->MailVmsStatus));
      return (FALSE);
   }

   InetMailParseHeader (vmptr);

   if (!vmptr->LooksLikeRfc)
   {
      StatusMessage (FI_LI, 1, "%s", LangFor("not_applicable"));
      return (FALSE);
   }

   CgiLibResponseHeader (200, "text/plain");

   if (vmptr->RfcHeaderPtr)
   {
      ch = vmptr->RfcHeaderPtr[vmptr->RfcHeaderLength];
      vmptr->RfcHeaderPtr[vmptr->RfcHeaderLength] = '\0';
      fputs (vmptr->RfcHeaderPtr, stdout);
      vmptr->RfcHeaderPtr[vmptr->RfcHeaderLength] = ch;
   }
   else
      fputs (vmptr->BodyPtr, stdout);

   return (TRUE);
}

/*****************************************************************************/
/*
Looks through each of (any) MIME parts for a likely message text.  Likely parts
are the first "text/plain" of a multipart without a file name, or a
non-multipart "text/plain" body.  If the message is not MIME then allocate and
populate a MIME structure using the textual body.

The 'ReadPreference' parameter provides the OPTIONS_READ_.. constants that
specify whether plain or HTML text is the preference.

Return a pointer to the MIME data representing the message body, NULL if there
is a problem.
*/

MIME_DATA* MessageBodyIdentify 
(
VMS_MAIL_MSG *vmptr,
int ReadPreference
)
{
   char  *cptr;
   MIME_DATA  *mdptr,
              *HtmlMimePtr,
              *PlainMimePtr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageBodyIdentify() !UL", ReadPreference);

   if (vmptr->VmsForeign) return (NULL);

   HtmlMimePtr = PlainMimePtr = NULL;

   mdptr = vmptr->MimeDataPtr;
   if (mdptr)
   {
      if (mdptr->ContentTypeIsMultipart)
      {
         /* a MIME message with a multipart content-type */
         for (mdptr = vmptr->MimeDataPtr; mdptr; mdptr = mdptr->NextMimePtr)
         {
            /* get the first instance of each of these */
            if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_PLAIN)
            {
               /* looking for plain text part without a file name */
               if (!PlainMimePtr)
                  if (!mdptr->NamePtr && !mdptr->FileNamePtr)
                     PlainMimePtr = mdptr;
            }
            else
            if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
            {
               /* looking for HTML text part without a file name */
               if (!HtmlMimePtr)
                  if (!mdptr->NamePtr && !mdptr->FileNamePtr)
                     HtmlMimePtr = mdptr;
            }
         }
      }
      else
      {
         /* must be in-line content */
         if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_PLAIN)
            PlainMimePtr = mdptr;
         else
         if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
            HtmlMimePtr = mdptr;
      }
   }
   else
   {
      /* just a plain-text message, fudge a MIME structure */
      mdptr = calloc (1, sizeof(MIME_DATA));
      if (!mdptr) ErrorExit (vaxc$errno, FI_LI);

      mdptr->TransferEncodingPtr = "8bit";
      mdptr->TransferEncodingIs = MIME_TRANSENC_8BIT;
      mdptr->ContentTypePtr = "text/plain";
      mdptr->ContentTypeIs = MIME_CONTENT_TEXT_PLAIN;
      if (vmptr->RfcBodyPtr)
      {
         /* RFC mail */
         mdptr->PartContentPtr = vmptr->RfcBodyPtr;
         mdptr->PartContentLength = vmptr->RfcBodyLength;
      }
      else
      {
         /* vanilla VMS Mail */
         mdptr->PartContentPtr = vmptr->BodyPtr;
         mdptr->PartContentLength = vmptr->BodyLength;
      }

      PlainMimePtr = mdptr;
   }

   if (vmptr->MimeDataPtr)
   {
      if (vmptr->MimeDataPtr->HtmlToPlainMimePtr)
      {
         /* a plain-text from HTML part has already been manufactured */
         PlainMimePtr = vmptr->MimeDataPtr->HtmlToPlainMimePtr;
      }
      else
      if (HtmlMimePtr && !PlainMimePtr)
      {
         /* there is an HTML but no plain-text part, manufacture one! */
         if (MimeDecProcessPart (HtmlMimePtr))
         {
            mdptr = (MIME_DATA*)calloc (1, sizeof(MIME_DATA));
            if (!mdptr) ErrorExit (vaxc$errno, FI_LI);

            mdptr->PartContentPtr = calloc (1, HtmlMimePtr->PartContentLength);
            if (!mdptr->PartContentPtr ) ErrorExit (vaxc$errno, FI_LI);
            mdptr->PartContentLength = HtmlMimePtr->PartContentLength;
            memcpy (mdptr->PartContentPtr,
                    HtmlMimePtr->PartContentPtr,
                    HtmlMimePtr->PartContentLength);
            mdptr->PartContentPtr[mdptr->PartContentLength] = '\0';
            mdptr->PartContentLength = HtmlToPlain (mdptr->PartContentPtr);
            mdptr->ContentTypeIs = MIME_CONTENT_TEXT_PLAIN;
            mdptr->ContentTypePtr = "text/plain";
            mdptr->TransferEncodingIs = MIME_TRANSENC_8BIT;
            mdptr->TransferEncodingPtr = "8bit";

            /* point the primary MIME structure to this manufactured part */
            PlainMimePtr = vmptr->MimeDataPtr->HtmlToPlainMimePtr = mdptr;
         }
      }
   }

   mdptr = NULL;
   if (ReadPreference == OPTIONS_READ_PLAIN1_HTML2)
   {
      if (PlainMimePtr)
         mdptr = PlainMimePtr;
      else
         mdptr = HtmlMimePtr;
   }
   else
   if (ReadPreference == OPTIONS_READ_PLAIN2_HTML1)
   {
      if (HtmlMimePtr)
         mdptr = HtmlMimePtr;
      else
         mdptr = PlainMimePtr;
   }
   else
   if (ReadPreference == OPTIONS_READ_PLAIN0_HTML1)
   {
      /* only used to 'probe' whether an HTML alternative is present! */
      if (HtmlMimePtr) mdptr = HtmlMimePtr;
   }
   else
   {
      /* anything else, including zero (not set), gets any plain text */
      mdptr = PlainMimePtr;
   }

   if (WatchEnabled)
   {
      if (mdptr)
         WatchThis ("BODY !AZ", mdptr->ContentTypePtr);
      else
         WatchThis ("BODY not found");
   }

   return (mdptr);
}

/*****************************************************************************/
/*
Make a message (ostensibly) more readable by ...

  1) Remove any PGP signature
  2) Collapse extra line breaks
  3) Wrap lines longer than specified by the macro SOY_WRAP_AT

All these are performed on the text in situ!
*/

int MessageMassage (char *TextPtr)

{
   int  colnum;
   char  *cptr, *sptr, *tptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageMassage()");

   MessageRemovePgpSig (TextPtr);

   MessageCollapseLineBreaks (TextPtr, 2);

   colnum = 0;
   cptr = sptr = TextPtr;
   while (*cptr)
   {
      if (*cptr == '\n')
      {
         colnum = 0;
         cptr++;
         continue;
      }
      if (colnum++ < SOY_WRAP_AT)
      {
         cptr++;
         continue;
      }
      for (tptr = cptr; tptr >= sptr; tptr--)
         if (isspace(*tptr)) break;
      if (tptr >= sptr)
      {
         /* found an intervening space */
         if (*(tptr+1) == '\n')
            tptr++;
         else
            *tptr++ = '\n';
         cptr = tptr;
         colnum = 0;
      }
      else
      {
         /* no intervening space, find next */
         while (*cptr && !isspace(*cptr))
         {
            colnum++;
            cptr++;
         }
      }
      sptr = cptr;
   }

   return (cptr-TextPtr);
}

/*****************************************************************************/
/*
Eliminate all empty lines and then collapse the number of consecutive newlines
down to whatevers specified in the parameter.
*/

int MessageCollapseLineBreaks
(
char *StringPtr,
int MaxConsecutive
)
{
   int  nlcnt;
   char  *cptr, *sptr, *solptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled)
      WatchThis ("MessageCollapseLineBreaks() !UL !AZ",
                 MaxConsecutive, StringPtr);

   if (MaxConsecutive < 1) MaxConsecutive = 1;
   cptr = sptr = StringPtr;
   while (*cptr)
   {
      solptr = sptr;
      while (*cptr && isspace(*cptr) && *cptr != '\n') *sptr++ = *cptr++;
      if (*cptr == '\n')
         sptr = solptr;
      else
         while (*cptr && *cptr != '\n') *sptr++ = *cptr++;
      while (*cptr == '\n') *sptr++ = *cptr++;
   }
   *sptr = '\0';

   cptr = sptr = StringPtr;
   while (*cptr == '\n') cptr++;
   while (*cptr)
   {
      while (*cptr && *cptr != '\n') *sptr++ = *cptr++;
      nlcnt = MaxConsecutive;
      while (*cptr == '\n' && nlcnt--) *sptr++ = *cptr++;
      while (*cptr == '\n') cptr++;
   }
   while (sptr > StringPtr && *(sptr-1) == '\n') sptr--;
   *sptr = '\0';

   return (sptr-StringPtr);
}

/*****************************************************************************/
/*
Return true if text appears PGP clear-text signed (RFC2440).
*/

BOOL MessagePgpSigned (char *TextPtr)

{
   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessagePgpSigned()");

   if (!TextPtr) return (FALSE);
   if (memcmp (TextPtr, "-----BEGIN PGP ", 15)) return (FALSE);
   return (TRUE);
}

/*****************************************************************************/
/*
Remove any PGP clear-text signature (RFC2440) from a message part.
*/

int MessageRemovePgpSig (char *TextPtr)

{
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageRemovePgpSig()");

   /* return if it doesn't appear to be PGP signed text */
   if (memcmp (TextPtr, "-----BEGIN PGP ", 15)) return (strlen(TextPtr));

   /* find the end-of-armor header empty line */
   for (cptr = sptr = TextPtr; *cptr; cptr++)
      if (*(unsigned short*)cptr == '\n\n' ||
          *(unsigned long*)cptr == '\r\n\r\n') break;

   if (!*cptr) return (strlen(TextPtr));

   if (*(unsigned short*)cptr == '\n\n')
      cptr += 2;
   else
      cptr += 4;

   while (*cptr)
   {
      while (*cptr && *cptr != '\n') *sptr++ = *cptr++;
      if (!*cptr) break;
      while (*cptr == '\n') *sptr++ = *cptr++;
      if (!*cptr) break;
      /* start of line */
      if (*cptr == '-')
      {
         if (*(unsigned short*)cptr == '- ')
            cptr += 2;
         else
         if (!memcmp (cptr, "-----BEGIN PGP ", 15))
            break;
      }
   }
   *sptr = '\0';

   return (sptr - TextPtr);
}

/*****************************************************************************/
/*
Return number of newline characters (lines) in text.
*/

int MessageLineCount (char *TextPtr)

{
   int  LineCount = 0;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageLineCount()");

   if (!TextPtr) return (0);

   for (cptr = TextPtr; *cptr; cptr++)
      if (*cptr == '\n') LineCount++;
   if (MessagePgpSigned (TextPtr)) LineCount -= 10;
   if (LineCount < 0) LineCount = 0;

   return (LineCount);
}

/*****************************************************************************/
/*
Retrieve the content of a specific MIME part (often an attachment) and provide
it to the client using the appropriate content-type.  If an HTML part then
sanitise the HTML unless it's in a popup (a full window of it's own) and the
user has set the appropriate configuration option.
*/

BOOL MessagePart (REQUEST_DATA *rdptr)

{
   BOOL  SanitiseHtml = FALSE;
   int  status,
        DataLength,
        PartCount;
   char  *cptr,
         *DataPtr;
   char  PartTypeField [64];
   MIME_DATA  *mdptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessagePart()");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;
   uoptr = &rdptr->UserOptions;

   vmptr->MessageId = rdptr->MessageId;
   vmptr->ConfirmHash = rdptr->MessageHash;

   CallMailMessageGet (muptr, NULL);
   if (VMSnok (muptr->MailVmsStatus))
   {
      if (muptr->MailVmsStatus == SS$_BADCHECKSUM ||
          muptr->MailVmsStatus == MAIL$_NOMOREMSG)
         StatusMessage (FI_LI, 1, "%s", LangFor("message_checksum"));
      else
         StatusMessage (FI_LI, 1, "MAIL: %s.", SysGetMsg(muptr->MailVmsStatus));
      return (FALSE);
   }

   InetMailParseHeader (vmptr);
   if (vmptr->MimeDataPtr)
      if (!MimeDecParseBody (vmptr->MimeDataPtr))
         vmptr->MimeDataPtr = NULL;

   if (vmptr->VmsForeign)
   {
      /* special case, VMS/FOREIGN message body */
      CgiLibResponseHeader (200, "application/octet-stream",
                            "Content-Length: %d\n", vmptr->BodyLength);
      if (CgiLibEnvironmentIsPurveyor())
      {
         CgiLibEnvironmentBinaryOut ();
         fwrite (vmptr->BodyPtr, vmptr->BodyLength, 1, stdout);
         CgiLibEnvironmentRecordOut ();
      }
      else
         fwrite (vmptr->BodyPtr, vmptr->BodyLength, 1, stdout);
      return (TRUE);
   }

   cptr = rdptr->AttachmentName;
   if (*(unsigned short*)cptr == '--')  /* hyphens */
   {
      SanitiseHtml = TRUE;
      cptr += 2;
   }

   PartCount = 0;
   for (mdptr = vmptr->MimeDataPtr; mdptr; mdptr = mdptr->NextMimePtr)
   {
      if (!mdptr->ContentTypeIsMultipart)
      {
         PartCount++;
         if (WatchEnabled)
            WatchThis ("!UL |!AZ|!AZ",
                       PartCount,
                       mdptr->NamePtr ? mdptr->NamePtr : "(null)",
                       mdptr->PartName);
         if (mdptr->NamePtr && strsame (mdptr->NamePtr, cptr, -1)) break;
         if (mdptr->PartName[0] && strsame (mdptr->PartName, cptr, -1)) break;
      }
   }

   if (mdptr)
   {
      /* part found */
      if (MimeDecProcessPart (mdptr))
      {
         DataPtr = mdptr->PartContentPtr;
         DataLength = mdptr->PartContentLength;

         if (!mdptr->ContentTypePtr) ErrorExit (SS$_BUGCHECK, FI_LI);
         /* use the content-type from the form */
         sprintf (PartTypeField, "FORM_PART_TYPE_%d", PartCount);
         CGIVARNULL (cptr, PartTypeField);
         if (!cptr)
            if (!*(cptr = rdptr->CgiQueryStringPtr))
               cptr = mdptr->ContentTypePtr;
         if (!*cptr || !isalpha(*cptr)) cptr = "application/octet-stream";
      }
   }
   else
   {
      /* part not found */
      if (*(unsigned short*)cptr == '__') cptr += 2;  /* underscores */
      /* the underscores virtually ensure [^] avoids part name clashes */
      if (LangSame ("attach_message", cptr))
      {
         /* the [full message] ("message/rfc822") button has been hit */
         DataPtr = vmptr->BodyPtr;
         DataLength = vmptr->BodyLength;

         CGIVARNULL (cptr, "FORM_PART_TYPE_0");
         if (!cptr)
            if (!*(cptr = rdptr->CgiQueryStringPtr))
               cptr = "message/rfc822";
         if (!*cptr || !isalpha(*cptr)) cptr = "application/octet-stream";
      }
      else
      {
         /* can't do any more */
         StatusMessage (FI_LI, 1, LangFor("part_not_found"));
         return (FALSE);
      }
   }

   if (mdptr && mdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
   {
      if (!cptr || !strncmp (cptr, "text/", 5))
      {
         if (uoptr->MassageRead) DataLength = MessageRemovePgpSig (DataPtr);
         if (SanitiseHtml) DataLength = HtmlSanitise (DataPtr);
      }
   }

   CgiLibResponseHeader (200, cptr, "Content-Length: %d\n", DataLength);

   CGIVARNULL (cptr, "FORM_PART_BTN");

   if (CgiLibEnvironmentIsPurveyor())
   {
      CgiLibEnvironmentBinaryOut ();
      fwrite (DataPtr, DataLength, 1, stdout);
      CgiLibEnvironmentRecordOut ();
   }
   else
      fwrite (DataPtr, DataLength, 1, stdout);

   return (TRUE);
}

/*****************************************************************************/
/*
Provide an HTML form selection list of the available message ranges.
*/

void MessageRangeSelect (REQUEST_DATA *rdptr)

{
   int  first, last,
        FirstId, LastId, ScratchId;
   char  *MsgRangeListPtr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageRangeSelect()");

   muptr = &VmsMailUser;

   if (rdptr->FolderMessageCount <= rdptr->UserOptions.MessagesPerPage)
   {
      fprintf (stdout, "<TD><!-- no message range --></TD>\n\n");
      return;
   }

   if (rdptr->SearchData.InProgress)
   {
      /* get highest and lowest id from this last search */
      LastId = rdptr->SearchData.LastIdSearched;
      FirstId = rdptr->SearchData.FirstIdSearched;
   }
   else
   if (rdptr->MsgIdListPtr)
   {
      /* get highest and lowest id in the current message list */
      LastId = atol(rdptr->MsgIdListPtr);
      FirstId = MessageRangeString (-1, rdptr->MsgIdListPtr);
   }
   else
      FirstId = LastId = 0;

   if (FirstId > LastId)
   {
      /* can happen when using the [prev][next] buttons */
      ScratchId = LastId;
      LastId = FirstId;
      FirstId = ScratchId;
   }
   if (WatchEnabled) WatchThis ("first:!UL last:!UL", FirstId, LastId);

   fprintf (stdout,
"<!-- message range begin -->\n\
<TD STYLE=\"text-align:left;white-space:nowrap;\">\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"range_prev\" VALUE=\"%s\"%s>&nbsp;\n",
            LangFor("prev"),
            CLICK_NOMORE_STARTWORKING(FirstId == 1));

   if (rdptr->SearchData.InProgress)
   {
      /* more of a reminder that we're searching than anything else */
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"main_btn\" VALUE=\"%s\">&nbsp;\n",
               LangFor("search_new"));
   }
   else
   {
      fprintf (stdout,
"<INPUT TYPE=\"hidden\" NAME=\"range_list_change\" VALUE=\"\">\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"range_msgs\" VALUE=\"%s\">\n\
<SELECT CLASS=\"css_select_1\" NAME=\"range_list\" \
ONCHANGE=\"range_list_change.value=true;startWorking();submit();\">\n\
<OPTION VALUE=\"%d\">&nbsp;%s\n",
               LangFor("range"),
               RANGE_MOST_RECENT,
               LangFor("most_recent"));

      last = rdptr->FolderMessageCount;
      while (last > 0)
      {
         first = last - rdptr->UserOptions.MessagesPerPage;
         if (first < 0) first = 0;
         first++;

         /* depending on which way we're moving */
         if (first == rdptr->RangeIdStart || last == rdptr->RangeIdStart)
            fprintf (stdout, "<OPTION VALUE=\"%d-%d\" SELECTED>&nbsp;%d-%d\n",
                     first, last, first, last);
         else
            fprintf (stdout, "<OPTION VALUE=\"%d-%d\">&nbsp;%d-%d\n",
                     first, last, first, last);

         last = first - 1;
      }

      fprintf (stdout,
"<OPTION VALUE=\"%d\"%s>&nbsp;%s\n\
</SELECT>&nbsp;\n",
               RANGE_ALL_MSGS,
               rdptr->RangeIdStart == RANGE_ALL_MSGS ? " SELECTED" : "",
               LangFor("all"));
   }

   fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"range_next\" VALUE=\"%s\"%s>\n\
</TD>\n\
<!-- message range end -->\n\n",
            LangFor("next"),
            CLICK_NOMORE_STARTWORKING(LastId == rdptr->FolderMessageCount));
}

/*****************************************************************************/
/*
Generate the list of messages.
*/

void MessageListInit (REQUEST_DATA *rdptr)

{
   int  cnt, status;
   unsigned long  EndSeconds,
                  StartSeconds;
   char  *cptr, *sptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_MSG  *vmptr, *vmlptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled)
      WatchThis ("MessageListInit() !SL !SL !SL",
                 rdptr->RangeIdStart, rdptr->RangeIdCount,
                 rdptr->RangeIdVector);

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;
   uoptr = &rdptr->UserOptions;

   if (VMSnok (muptr->MailVmsStatus)) return;

   muptr->ListMessageIdStart = rdptr->RangeIdStart;
   muptr->ListMessageIdCount = rdptr->RangeIdCount;
   muptr->ListMessageIdVector = rdptr->RangeIdVector;

   if (rdptr->SearchData.InProgress)
   {
      muptr->SearchFunction = &SearchMessage;
      muptr->SearchDataPtr = (void*)&rdptr->SearchData;
      muptr->ListMessageIdMax = rdptr->SearchData.ResultsEvery;
      if (SoyMailConfig.SelectNotSearch)
         muptr->SelectNotSearch = SoyMailConfig.SelectNotSearch;
      else
         muptr->SelectNotSearch = rdptr->SearchData.SelectNotSearch;
      rdptr->SearchData.SearchCount =
         rdptr->SearchData.HitCount =
         rdptr->SearchData.FirstIdSearched =
         rdptr->SearchData.LastIdSearched = 0;
      time (&StartSeconds);
   }
   else
      rdptr->SearchData.SearchSeconds = 0;

   CallMailMessageList (muptr);

   if (rdptr->SearchData.InProgress)
   {
      time (&EndSeconds);
      rdptr->SearchData.SearchSeconds = EndSeconds - StartSeconds;
      /* separately otherwise DEC C V5.7-004 OpenVMS VAX V6.0 complains */
      muptr->SearchFunction = NULL;
      muptr->SearchDataPtr = NULL;
      muptr->ListMessageIdMax = 0;
      /* restore if the process priority was changed during the search */
      SearchRestorePriority ();
   }

   if (Debug)
      fprintf (stdout, "%%X%08.08X MessageListCount:%d\n",
               muptr->MailVmsStatus, muptr->MessageListCount);

   if (VMSnok (muptr->MailVmsStatus))
   {
      if (muptr->MailVmsStatus == MAIL$_NOTEXIST)
         if (LangSame ("(none)", rdptr->FolderName))
            StatusMessage (FI_LI, 0, LangFor("folder_select"));
         else
            StatusMessage (FI_LI, 0, LangFor("folder_empty"));
      else
         StatusMessage (FI_LI, 1, "MAIL: %s.",
                        SysGetMsg(muptr->MailVmsStatus));

      /* not a razoo */
      rdptr->RangeIdStart = rdptr->RangeIdCount =
         rdptr->FolderMessageCount = 0;
   }
   else
   {
      if (rdptr->PrivateAccess)
         cptr = FolderDescribe (rdptr->MailFileName, rdptr->FolderName);
      else
      if (rdptr->PublicWildcard)
         cptr = HTML_ESCAPE(rdptr->FolderName);
      else
         cptr = rdptr->PublicAccessString;

      if (rdptr->SearchData.InProgress)
      {
         char  SearchMessage [512];
         sptr = SearchMessage;
         sptr += sprintf (sptr, LangFor("search_searched"),
                          HTML_ESCAPE(cptr),
                          vmptr->MessageSelectedCount,
                          rdptr->SearchData.SearchCount,
                          HTML_ESCAPE(rdptr->SearchData.Criteria));
         if (rdptr->SearchData.SearchSeconds >= 10)
            sprintf (sptr, "&nbsp; %s: %02.02d:%02.02d.",
                     LangFor("duration"),
                     rdptr->SearchData.SearchSeconds / 60,
                     rdptr->SearchData.SearchSeconds % 60);
         StatusMessage (FI_LI, 0, "%s", SearchMessage);
      }
      else
         StatusMessage (FI_LI, 0, LangFor("folder_opened"),
                        HTML_ESCAPE(cptr), vmptr->MessageSelectedCount);

      /* how many messages are in the folder */
      rdptr->FolderMessageCount = vmptr->MessageSelectedCount;
   }

   if (rdptr->MsgIdListPtr)
   {
      /* it is also used to store checkbox numbers */
      free (rdptr->MsgIdListPtr);
      rdptr->MsgIdListPtr = NULL;
   }
   rdptr->MsgIdListCount = muptr->MessageListCount;
   if (rdptr->MsgIdListCount)
   {
      rdptr->MsgIdListPtr = sptr = calloc (rdptr->MsgIdListCount, 32);
      if (!rdptr->MsgIdListPtr) ErrorExit (vaxc$errno, FI_LI);
      cptr = "";
      for (vmlptr = muptr->MessageList; vmlptr; vmlptr = vmlptr->NextPtr)
      {
         sptr += sprintf (sptr, "%s%d.%d",
                          cptr, vmlptr->CurrentId, vmlptr->MessageHash);
         cptr = ",";
      }
   }
   else
      rdptr->MsgIdListPtr = "";
}

/*****************************************************************************/
/*
Display the basic message folder and message list page.
*/

void MessageFolderList (REQUEST_DATA *rdptr)

{
   int  cnt, idx, status;

   char  *cptr, *sptr, *zptr;
   char  FromBuffer [256],
         MessageFlag [96],
         SubjBuffer [256];
   USER_OPTIONS  *uoptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("MessageFolderList()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;
   vmptr = &muptr->VmsMailMsg;

   fprintf (stdout,
"<P CLASS=\"css_p08\">\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"6\" \
CLASS=\"css_panel_1\"><TR>\n\
\n");

   if (uoptr->FolderButtons)
   {
      fprintf (stdout,
"<TD><TABLE WIDTH=\"100%%\" CELLPADDING=\"0\" CELLSPACING=\"0\" BORDER=\"0\">\n\
\n\
<!-- folder buttons begin -->\n\
<TR><TD WIDTH=\"5%%\" ALIGN=\"left\" VALIGN=\"top\" \
STYLE=\"padding:0.3em;padding-right:1.2em;\
font-size:0.8em;white-space:nowrap;\">\n");

      FolderButtonListEach (rdptr);

      fprintf (stdout,
"</TD><TD WIDTH=\"95%%\" ALIGN=\"left\" VALIGN=\"top\">\n\
<!-- folder buttons end -->\n\
\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"2\" CELLSPACING=\"0\" BORDER=\"0\"><TR>\n\n");
   }

   /******************/
   /* message ranges */
   /******************/

   MessageRangeSelect (rdptr);

   /**************************/
   /* message action buttons */
   /**************************/

   fprintf (stdout,
"<!-- message action begin -->\n\
<SCRIPT LANGUAGE=\"JavaScript\">\n\
function checkAllMsg (action) {\n\
   if (!document.soymail_form.msg_cbx_all.checked) {\n\
      startWorking();\n\
      return true;\n\
   }\n\
   var prompt = action + \" %s %s. \" + \"%s\";\n\
   var ok = confirm (prompt);\n\
   if (ok) startWorking();\n\
   return ok;\n\
}\n\
</SCRIPT>\n\
\
<TD STYLE=\"text-align:right;vertical-align:top;white-space:nowrap;\">\n\
<SPAN CLASS=\"css_nw\">\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_copy\" VALUE=\"%s &gt;&gt;\" \
ONCLICK=\"return checkAllMsg(\'%s\')\">\n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_move\" VALUE=\"%s &gt;&gt;\" \
ONCLICK=\"return checkAllMsg(\'%s\')\">\n\
<SELECT CLASS=\"css_select_1\" NAME=\"folder_list_dest\" \
ONCHANGE=\"document.soymail_form.folder_text_dest.value=\'\'\">\n",
            LangFor("all"), LangFor("all_messages"), LangFor("sure?"),
            LangFor("copy"), LangFor("copy"),
            LangFor("move"), LangFor("move"));

   FolderDestListEach (rdptr);

   CGIVAR (cptr, "FORM_FOLDER_TEXT_DEST");

   fprintf (stdout,
"</SELECT>\n\
<INPUT CLASS=\"css_text_1\" TYPE=\"text\" \
NAME=\"folder_text_dest\" SIZE=\"8\" VALUE=\"%s\" \
ONCLICK=\"this.select()\">&nbsp;\n\
</SPAN> <SPAN CLASS=\"css_nw\">\n",
            HTML_ESCAPE(cptr));

   if (!strcmp (rdptr->FolderName+1, LangFor("folder_soy_waste")))
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_empty\" VALUE=\"%s\" \
ONCLICK=\"return checkAllMsg(\'%s\')\">\n",
               LangFor("empty"), LangFor("empty"));
   else
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_waste\" VALUE=\"%s\" \
ONCLICK=\"return checkAllMsg(\'%s\')\">\n",
               LangFor("waste"), LangFor("waste"));

   fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_delete\" VALUE=\"%s\" \
ONCLICK=\"return checkAllMsg(\'%s\')\">\n\
</SPAN>\n\
</TD>\n\
</TR>\n\
<!-- message action end -->\n\n",
            LangFor("delete"), LangFor("delete"));

   if (!muptr->MessageListCount)
   {
      /****************/
      /* empty folder */
      /****************/

      rdptr->RangeIdStart = rdptr->RangeIdCount = 0;

      if (uoptr->FolderButtons)
         fprintf (stdout, "\n</TD></TR>\n</TABLE>\n\n</TD></TR>\n</TABLE>\n\n");

      fprintf (stdout, "</TABLE>\n");

      MessageListFooter ();

      fprintf (stdout,
"</TD></TR>\n\
</TABLE>\n\
<!-- message list end -->\n\n");

      return;
   }

   /********************/
   /* list of messages */
   /********************/

   if (rdptr->ClientHasJavaScript)
      fprintf (stdout,
"<!-- message list begin -->\n\
<SCRIPT LANGUAGE=\"JavaScript\">\n\
\
function checkAllId () {\n\
   for (idx = 0; idx < document.soymail_form.elements.length; idx++) {\n\
      if (document.soymail_form.elements[idx].type == \"checkbox\" &&\n\
          document.soymail_form.elements[idx].name.substring(0,8) \
== \"msg_cbx_\" &&\n\
          !isNaN(parseInt(document.soymail_form.\
elements[idx].name.substring(8,9))))\n\
         document.soymail_form.elements[idx].checked = \
document.soymail_form.msg_cbx_page.checked;\n\
   }\n\
}\n\
\
function targetPopup (pop) {\n\
   document.soymail_form.msg_btn_popup.value = pop;\n\
   if (pop) targetBlank();\n\
}\n\
</SCRIPT>\n");

   fprintf (stdout,
"<INPUT TYPE=\"hidden\" NAME=\"msg_btn_popup\" VALUE=\"\">\n\
<INPUT TYPE=\"hidden\" NAME=\"msg_range_list\" VALUE=\"%s\">\n\
<TR><TD COLSPAN=\"2\">\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"5\" BORDER=\"0\">\n\
<TR><TD>",
            rdptr->MsgIdListPtr);

   if (rdptr->ClientHasJavaScript)
   {
      /* careful how this is done or Opera produces a checkbox artifact */
      fprintf (stdout,
"<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT TYPE=\"checkbox\" NAME=\"msg_cbx_page\" VALUE=\"1\" \
ONCLICK=\"return checkAllId()\">\')\
</SCRIPT>");
   }

   fprintf (stdout,
"</TD><TD></TD><TD></TD>\
<TD CLASS=\"css_nwtar5\" STYLE=\"font-weight:bold;text-decoration:underline;\">%s</TD>\
<TD CLASS=\"css_nwtal5\" STYLE=\"font-weight:bold;text-decoration:underline;\">%s</TD>\
<TD CLASS=\"css_nwtal5\" STYLE=\"font-weight:bold;text-decoration:underline;\">%s</TD>\
</TR>\n",
            LangFor("received"),
            LangFor("from"),
            LangFor("subject"));

   for (vmptr = muptr->MessageList; vmptr; vmptr = vmptr->NextPtr)
   {
      if (Debug)
         fprintf (stdout,
"%%X%08.08X id:%d size:%d hash:%d\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n\n",
                  muptr->MailVmsStatus,
                  vmptr->CurrentId, vmptr->SizeInRecords, vmptr->MessageHash,
                  vmptr->Date, vmptr->From, vmptr->Subject,
                  vmptr->To, vmptr->Cc);

      zptr = (sptr = FromBuffer) + sizeof(FromBuffer)-1;
      for (cptr = vmptr->From; *cptr && sptr < zptr; *sptr++ = *cptr++);
      *sptr = '\0';
      MimeDecIsoString (FromBuffer);
      FromBuffer[40] = ']';
      FromBuffer[41] = '\0';

      zptr = (sptr = SubjBuffer) + sizeof(SubjBuffer)-1;
      for (cptr = vmptr->Subject; *cptr && sptr < zptr; *sptr++ = *cptr++);
      *sptr = '\0';
      MimeDecIsoString (SubjBuffer);
      SubjBuffer[60] = ']';
      SubjBuffer[61] = '\0';

      /* indicate whether this message recently has been read */
      strcpy (MessageFlag, "<TD></TD>");
      if (vmptr->ErrorOpenIn)
      {
         /* error opening an external mail file */
         strcpy (MessageFlag,
"<TD CLASS=\"css_warning\" \
STYLE=\"text-align:right;padding-right:0.2em;\">!</TD>");
      }
      else
      if (uoptr->MessageReadHistory)
      {
         for (idx = 0; idx < rdptr->MessageReadHistoryCount; idx++)
         {
            if (rdptr->MessageReadHistory[idx] != vmptr->CurrentId) continue;
            if (uoptr->MessageReadHistory == OPTIONS_READ_HISTORY_ASTER)
               strcpy (MessageFlag, "<TD CLASS=\"css_rhist\">*</TD>");
            else
               sprintf (MessageFlag, "<TD CLASS=\"css_rhist\">-%d</TD>", idx+1);
            break;
         }
      }

      /* this mix of styles to get both Mozilla and MSIE to look right! */
      fprintf (stdout,
"<TR CLASS=\"css_nw\">\
<TD><INPUT TYPE=\"checkbox\" NAME=\"msg_cbx_%d\" VALUE=\"1\"></TD>\
<TD>\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:3.5em;\" \
TYPE=\"submit\" NAME=\"msg_btn\" VALUE=\"%04d\" \
ONCLICK=\"targetPopup(0);\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_btn_%d\" VALUE=\"^\" \
ONCLICK=\"targetPopup(%d);\">\');\
</SCRIPT>\
</TD>\
%s\
<TD CLASS=\"css_nwtar5\">%17.17s</TD>\
<TD CLASS=\"css_nwtal5\">%s</TD>\
<TD CLASS=\"css_nwtal5\">%s</TD>\
</TR>\n",
         vmptr->CurrentId,
         vmptr->CurrentId,
         vmptr->CurrentId,
         vmptr->CurrentId,
         MessageFlag,
         InetMailRfcVmsDate(vmptr->Date),
         HTML_ADDR(FromBuffer),
         vmptr->Subject[0] ? HTML_TEXT(SubjBuffer) : LangFor("(none)"));
   }

   
   if (!rdptr->SearchData.InProgress)
      fprintf (stdout,
"<TR CLASS=\"css_nw\">\
<TD><INPUT TYPE=\"checkbox\" NAME=\"msg_cbx_all\" VALUE=\"1\"></TD>\
<TD COLSPAN=\"4\" STYLE=\"font-size:0.8em;\">%s %d %s</TD></TR>\n",
               LangFor("all"), rdptr->FolderMessageCount,
               LangFor("all_messages"));

   fprintf (stdout, "</TABLE>\n<!-- message list end -->\n\n");

   if (uoptr->FolderButtons)
      fprintf (stdout, "</TD></TR>\n</TABLE>\n\n</TD></TR>\n</TABLE>\n\n");

   fprintf (stdout, "</TD></TR>\n</TABLE>\n\n");

   MessageListFooter ();
}

/*****************************************************************************/
/*
*/

void MessageListFooter ()

{
   /*********/
   /* begin */
   /*********/

   if (!SoyMailConfig.MessageListFooter) return;

   fprintf (stdout,
"<!-- site footer begin -->\n\
<P CLASS=\"css_p08\">\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"5\" \
CLASS=\"css_panel_1\">\n\
<TR><TD ALIGN=\"left\" VALIGN=\"top\" STYLE=\"white-space:normal;\">\
%s\n\
</TD></TR>\n\
</TABLE>\n\
<!-- site footer end -->\n\n",
            SoyMailConfig.MessageListFooter);
}

/*****************************************************************************/
/*
Copy, with optional delete (i.e. a move), from one folder to another.
*/

void MessageCopy
(
REQUEST_DATA *rdptr,
BOOL DeleteAfterCopy
)
{
   int  cnt, idx;
   int  *midptr;
   char  *cptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   muptr = &VmsMailUser;

   if (WatchEnabled)
      WatchThis ("MessageCopy() !AZ !AZ/!AZ !AZ/!AZ !AZ",
                 DeleteAfterCopy ? "MOVE" : "COPY",
                 rdptr->MailFileName, rdptr->FolderName,
                 rdptr->DestMailFileName, rdptr->DestFolderName,
                 rdptr->MsgIdListPtr ? rdptr->MsgIdListPtr : "(null)");

   if (!strcmp (rdptr->MailFileName, rdptr->DestMailFileName))
   {
      if (!strcmp (muptr->VmsMailFolderName, rdptr->DestFolderName))
      {
         StatusMessage (FI_LI, 1, "%s", LangFor("folder_same"));
         return;
      }
   }

   if (rdptr->DestMailFileNameLength)
   {
      strcpy (muptr->VmsMailDestFileName, rdptr->DestMailFileName);
      muptr->VmsMailDestFileNameLength = rdptr->DestMailFileNameLength;
   }
   else
   {
      strcpy (muptr->VmsMailDestFileName, rdptr->MailFileName);
      muptr->VmsMailDestFileNameLength = rdptr->MailFileNameLength;
   }
   strcpy (muptr->VmsMailDestFolderName, rdptr->DestFolderName);
   muptr->VmsMailDestFolderNameLength = rdptr->DestFolderNameLength;

   muptr->MsgIdListPtr = rdptr->MsgIdListPtr;
   muptr->MsgIdListCount = rdptr->MsgIdListCount;

   if (!muptr->MsgIdListCount)
   {
      StatusMessage (FI_LI, 1, "%s", LangFor("none_selected"));
      return;
   }

   if (!muptr->VmsMailDestFolderNameLength)
   {
      StatusMessage (FI_LI, 1, "%s", LangFor("dest_not_specified"));
      return;
   }

   if (DeleteAfterCopy)
      CallMailMessageMove (muptr);
   else
      CallMailMessageCopy (muptr);
   if (Debug) fprintf (stdout, "%%X%08.08X\n", muptr->MailVmsStatus);

   if (VMSnok (muptr->MailVmsStatus))
   {
      if (muptr->MailVmsStatus == SS$_BADCHECKSUM ||
          muptr->MailVmsStatus == MAIL$_NOMOREMSG)
         StatusMessage (FI_LI, 1, "%s", LangFor("message_checksum"));
      else
         StatusMessage (FI_LI, 1, "MAIL: %s", SysGetMsg(muptr->MailVmsStatus));
   }
   else
   {
      /* just in case we've tinkered with NEWMAIL */
      rdptr->NewMessages = muptr->VmsMailNewMessages;

      cptr = FolderDescribe (rdptr->DestMailFileName, rdptr->DestFolderName);

      StatusMessage (FI_LI, 0, "%d %s &quot;%s&quot;.",
                     muptr->MessageCopiedCount,
                     DeleteAfterCopy ? LangFor("msgs_moved") :
                                       LangFor("msgs_copied"),
                     HTML_ESCAPE(cptr));
   }
}

/*****************************************************************************/
/*
Delete the specified messages from the folder.
*/

void MessageDelete (REQUEST_DATA *rdptr)

{
   int  cnt, idx;
   int  *midptr;
   char  *cptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   muptr = &VmsMailUser;

   if (WatchEnabled)
      WatchThis ("MessageDelete() !AZ",
                 rdptr->MsgIdListPtr ? rdptr->MsgIdListPtr : "(null)");

   strcpy (muptr->VmsMailFileName, rdptr->MailFileName);
   muptr->VmsMailFileNameLength = rdptr->MailFileNameLength;

   muptr->MsgIdListPtr = rdptr->MsgIdListPtr;
   muptr->MsgIdListCount = rdptr->MsgIdListCount;

   if (!muptr->MsgIdListCount)
   {
      StatusMessage (FI_LI, 1, "%s", LangFor("none_selected"));
      return;
   }

   CallMailMessageDelete (muptr);
   if (Debug)
      fprintf (stdout, "%%X%08.08X MessageDeletedCount:%d\n",
               muptr->MailVmsStatus, muptr->MessageDeletedCount);

   if (VMSnok (muptr->MailVmsStatus))
   {
      if (muptr->MailVmsStatus == SS$_BADCHECKSUM ||
          muptr->MailVmsStatus == MAIL$_NOMOREMSG)
         StatusMessage (FI_LI, 1, "%s", LangFor("message_checksum"));
      else
         StatusMessage (FI_LI, 1, "MAIL: %s.", SysGetMsg(muptr->MailVmsStatus));
   }
   else
   {
      /* just in case we've tinkered with NEWMAIL */
      rdptr->NewMessages = muptr->VmsMailNewMessages;

      cptr = FolderDescribe (rdptr->MailFileName, rdptr->FolderName);

      StatusMessage (FI_LI, 0, "%d %s &quot;%s&quot;.",
                     muptr->MessageDeletedCount,
                     LangFor("msgs_deleted"),
                     HTML_ESCAPE(cptr));
   }
}

/*****************************************************************************/

