/*****************************************************************************/
/*
                                  public.c

For public access ...

Displays a list of messages in a folder.
Displays the contents of a message (header, body, links to attachments).
Displays a 'print-view' (more suitable for printing) of a message.
Displays the raw message (header+body, or header alone).

Of course with limitations and in a format that is suitable for public access!


COPYRIGHT
---------
Copyright (C) 2005,2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
03-APR-2005  MGD  significant rework to rationalise public and private
                  significant changes to support public wildcard folders
30-MAR-2006  MGD  PublicMessagePage() precede each message part/attachment
                  with an anchor/link for right-click browser save-as
28-FEB-2006  MGD  major refinements to public access
01-FEB-2005  MGD  initial
*/

/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <mailmsgdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header file */
#include "soymail.h"
#include "attach.h"
#include "callmail.h"
#include "cgilib.h"
#include "compose.h"
#include "config.h"
#include "html.h"
#include "mainmenu.h"
#include "other.h"
#include "public.h"
#include "message.h"
#include "request.h"

#define FI_LI __FILE__, __LINE__

/* global storage */

/* external storage */

extern BOOL  Debug,
             WatchEnabled;

extern int  VmsVersion;

extern char  *CgiEnvironmentPtr,
             *SoyMailStatusPtr;

extern char  CurrentVmsTimeString[],
             SoftwareEnv[],
             SoftwareId[],
             SoftwareVn[];

extern CONFIG_DATA  SoyMailConfig;
extern REQUEST_DATA  RequestData;
extern VMS_MAIL_USER  VmsMailUser;

/* ad hoc prototypes */
int lib$day_of_week(__unknown_params);
int sys$fao(__unknown_params);

/****************************************************************************/
/*
Parse the buttons associated with the public folder message list.
*/ 

void PublicFolderRequest (REQUEST_DATA *rdptr)

{
   int  cnt, hash;
   char  *cptr, *sptr, *zptr,
         *MsgRangeListPtr;
   char  MsgBtnString [32],
         MsgCbxString [32],
         MsgHashString [32];
   VMS_MAIL_USER  *muptr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("PublicFolderRequest()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;
 
   if (uoptr->MassageRead) rdptr->MassageMessage = TRUE;

   /* the string containing comma-separated message id dot hashes */
   CGIVAR (MsgRangeListPtr, "FORM_MSG_RANGE_LIST");

   /*****************/
   /* search button */
   /*****************/

   CGIVARNULL (cptr, "FORM_MAIN_BTN");
   if (cptr && *cptr)
   {
      /* searching can be disabled */
      rdptr->SearchData.InProgress = FALSE;
      if (LangSame ("search", cptr) ||
          LangSame ("search_new", cptr))
      {
         if (SearchPageRequest (rdptr)) return;
      }
   }

   /*****************************/
   /* individual message button */
   /*****************************/

   CGIVARNULL (cptr, "FORM_MSG_BTN");
   if (cptr)
   {
      rdptr->MessageId = atol(cptr);
      rdptr->MessageHash = MessageRangeString (rdptr->MessageId,
                                               MsgRangeListPtr);
      PublicMessagePage (rdptr);
      return;
   }

   /************************/
   /* popup message button */
   /************************/

   CGIVARNULL (cptr, "FORM_MSG_BTN_POPUP");
   if (cptr && (cnt = atol(cptr)))
   {
      rdptr->MessageId = cnt;
      rdptr->MessageHash = MessageRangeString (rdptr->MessageId,
                                               MsgRangeListPtr);
      PublicMessagePage (rdptr);
      return;
   }

   /*****************/
   /* message range */
   /*****************/

   /* the range button or range 'onchange=' */
   CGIVARNULL (cptr, "FORM_RANGE_MSGS");
   if (!cptr) CGIVARNULL (cptr, "FORM_RANGE_LIST_CHANGE");
   if (cptr && *cptr) CGIVARNULL (cptr, "FORM_RANGE_LIST");
   /* also check that the range list is not empty */
   if (cptr && *cptr)
   {
      rdptr->RangeIdStart = atoi(cptr);
      while (*cptr && *cptr != '-') cptr++;
      if (*cptr)
      {
         cptr++;
         rdptr->RangeIdCount = atoi(cptr);
         rdptr->RangeIdCount -= rdptr->RangeIdStart - 1;
      }
      else
         rdptr->RangeIdCount = uoptr->MessagesPerPage;
      rdptr->RangeIdVector = +1;
      PublicFolderPage (rdptr);
      return;
   }

   /******************************/
   /* previous range of messages */
   /******************************/

   CGIVARNULL (cptr, "FORM_RANGE_PREV");
   if (cptr)
   {
      /* get the lowest id in the list and start one less than that */
      rdptr->RangeIdStart = MessageRangeString (-1, MsgRangeListPtr);
      if (rdptr->RangeIdStart <= 1)
      {
         rdptr->RangeIdStart = sdptr->RangeIdStart;
         StatusMessage (FI_LI, 1, LangFor("no_more_messages"));
      }
      else
         rdptr->RangeIdStart--;
      rdptr->RangeIdCount = uoptr->MessagesPerPage;
      rdptr->RangeIdVector = -1;
      PublicFolderPage (rdptr);
      return;
   }

   /**************************/
   /* next range of messages */
   /**************************/

   CGIVARNULL (cptr, "FORM_RANGE_NEXT");
   if (cptr)
   {
      /* the highest id is always the first one in the string */
      rdptr->RangeIdStart = atoi(MsgRangeListPtr) + 1;
      if (rdptr->RangeIdStart >= sdptr->FolderMessageCount)
      {
         /* just go back to the start of whatever we had before */
         rdptr->RangeIdStart = sdptr->RangeIdStart;
         StatusMessage (FI_LI, 1, LangFor("no_more_messages"));
      }
      rdptr->RangeIdCount = uoptr->MessagesPerPage;
      rdptr->RangeIdVector = +1;
      PublicFolderPage (rdptr);
      return;
   }

   /******************************/
   /* default to latest messages */
   /******************************/

   rdptr->RangeIdStart = RANGE_MOST_RECENT;
   rdptr->RangeIdCount = uoptr->MessagesPerPage;

   PublicFolderPage (rdptr);
}

/****************************************************************************/
/*
Parse the buttons from the public message display page.
*/ 

void PublicMessageRequest (REQUEST_DATA *rdptr)

{
   static char  SingleMessageList [32];

   char  *cptr, *sptr, *zptr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("PublicMessageRequest()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;

   if (uoptr->MassageRead) rdptr->MassageMessage = TRUE;


   if (rdptr->PublicWildcard)
   {
      /*******************/
      /* wildcard public */
      /*******************/

      CGIVARNULL (cptr, "FORM_FOLDER_LIST_CHANGE");
      if (cptr && *cptr == 't')
      {
         PublicFolderRequest (rdptr);
         return;
      }

      CGIVARNULL (cptr, "FORM_MAIN_BTN");
      if (cptr && *cptr)
      {
         rdptr->SearchData.InProgress = FALSE;
         if (LangSame ("open", cptr))
         {
            PublicFolderRequest (rdptr);
            return;
         }
      }
   }

   /****************************/
   /* previous/at/next message */
   /****************************/

   CGIVARNULL (cptr, "FORM_MSG_PREV");
   if (cptr)
   {
      rdptr->MessageId = sdptr->MessageId;
      if (rdptr->MessageId > 1)
         rdptr->MessageId--;
      else
         StatusMessage (FI_LI, 1, LangFor("no_more_messages"));
      rdptr->MessageHash = 0;
      PublicMessagePage (rdptr);
      return;
   }

   CGIVARNULL (cptr, "FORM_MSG_FOLDER_AT");
   if (cptr)
   {
      rdptr->RangeIdCount = uoptr->MessagesPerPage;
      for (rdptr->RangeIdStart = sdptr->FolderMessageCount;
           rdptr->RangeIdStart >= sdptr->MessageId;
           rdptr->RangeIdStart -= rdptr->RangeIdCount);
      rdptr->RangeIdStart++;
      if (rdptr->RangeIdStart < 1) rdptr->RangeIdStart = 1;
      PublicFolderPage (rdptr);
      return;
   }

   CGIVARNULL (cptr, "FORM_MSG_NEXT");
   if (cptr)
   {
      rdptr->MessageId = sdptr->MessageId;
      if (rdptr->MessageId < sdptr->FolderMessageCount)
         rdptr->MessageId++;
      else
         StatusMessage (FI_LI, 1, LangFor("no_more_messages"));
      rdptr->MessageHash = 0;
      PublicMessagePage (rdptr);
      return;
   }

   /********************/
   /* specific message */
   /********************/

   CGIVARNULL (cptr, "FORM_MSG_ID_HASH");
   if (cptr)
   {
      rdptr->MessageId = atol(cptr);
      while (*cptr && *cptr != '.') cptr++;
      if (*cptr) cptr++;
      if (!*cptr) ErrorExit (SS$_BUGCHECK, FI_LI);;
      rdptr->MessageHash = atol(cptr);
   }

   /*****************/
   /* message parts */
   /*****************/

   CGIVARNULL (cptr, "FORM_PART_BTN");
   if (cptr)
   {
      zptr = (sptr = rdptr->AttachmentName) + sizeof(rdptr->AttachmentName)-1;
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
      *sptr = '\0';
      /* if the attachment is successfully accessed then just finish up */
      if (MessagePart (rdptr)) return;
      PublicMessagePage (rdptr);
      return;
   }

   /************************/
   /* message read buttons */
   /************************/

   CGIVARNULL (cptr, "FORM_MSG_PRINT");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_PRINT_POPUP");
   if (cptr)
   {
      if (MessagePrintView (rdptr)) return;
      PublicMessagePage (rdptr);
      return;
   }

   CGIVARNULL (cptr, "FORM_MSG_RAW");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_RAW_POPUP");
   if (cptr)
   {
      if (MessageRawView (rdptr)) return;
      PublicMessagePage (rdptr);
      return;
   }

   CGIVARNULL (cptr, "FORM_MSG_HDR");
   if (!cptr) CGIVARNULL (cptr, "FORM_MSG_HDR_POPUP");
   if (cptr)
   {
      if (MessageHdrView (rdptr)) return;
      PublicMessagePage (rdptr);
      return;
   }

   /**************/
   /* HTML/plain */
   /**************/

   CGIVARNULL (cptr, "FORM_MSG_READ_PREF");
   if (cptr)
   {
      /* one is on, minus one is off, zero is not set! */
      if (LangSame ("message_read_html", cptr))
         rdptr->MessageReadPreference = OPTIONS_READ_PLAIN2_HTML1;
      if (LangSame ("message_read_plain", cptr))
         rdptr->MessageReadPreference = OPTIONS_READ_PLAIN1_HTML2;
      PublicMessagePage (rdptr);
      return;
   }

   /****************/
   /* message wrap */
   /****************/

   CGIVARNULL (cptr, "FORM_MSG_WRAP");
   if (cptr)
   {
      /* one is on, minus one is off, zero is not set! */
      if (LangSame ("message_massage_on", cptr))
         rdptr->MassageMessage = TRUE;
      else
      if (LangSame ("message_massage_off", cptr))
         rdptr->MassageMessage = FALSE;
      /* use the read preference from the previous request */
      rdptr->MessageReadPreference = sdptr->MessageReadPreference;
      PublicMessagePage (rdptr);
      return;
   }

   /*********************/
   /* message ID in URL */
   /*********************/

   PublicMessagePage (rdptr);
}

/*****************************************************************************/
/*
Generate the HTML for a public page (folder message list or message display).
*/

void PublicPageBegin
(
REQUEST_DATA *rdptr,
int ThisPageIdent
)
{
   int  NewMessages;
   char  *cptr,
         *UserNamePtr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("PublicPageBegin()");

   uoptr = &rdptr->UserOptions;
   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;

   if (!rdptr->ThisPageIdent) rdptr->ThisPageIdent = ThisPageIdent;

   if (rdptr->CharSetPtr) CgiLibResponseSetCharset (rdptr->CharSetPtr);

   CgiLibResponseHeader (200, "text/html");

   fprintf (stdout,
"<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" VALUE=\"%s\">\n\
<META NAME=\"environment\" VALUE=\"%s\">\n\
<META NAME=\"language\" VALUE=\"%s %s\">\n",
            SoftwareEnv,
            CgiEnvironmentPtr,
            LangFor("lang_name"),
            LangFor("lang_author"));

   /* used to support some of the enhanced functionality */
   SoyMailJavaScriptStuff (rdptr);

   SoyMailStyleStuff (rdptr, uoptr->Theme);

   fprintf (stdout,
"<TITLE>%s</TITLE>\n\
</HEAD>\n\
\n\
<BODY ONLOAD=\"refreshClock();\" \
ONUNLOAD=\"clearWorking();\" \
ONFOCUS=\"refreshSoyMailForm();\" ONBLUR=\"refreshSoyMailForm();\">\n\
<FORM NAME=\"soymail_form\" ACTION=\"%s\" \
METHOD=\"POST\" ENCTYPE=\"multipart/form-data\">\n\
\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"0\" CELLSPACING=\"5\" ALIGN=\"center\">\n\
<TR><TD>\n\n",
            SoyMailPageTitle(rdptr),
            rdptr->FormAction);
}

/*****************************************************************************/
/*
*/

void PublicPageEnd (REQUEST_DATA *rdptr)

{
   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("PublicPageEnd()");

   fprintf (stdout,
"</TD></TR>\n\
</TABLE>\n\n\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT TYPE=\"hidden\" \
NAME=\"javascript\" VALUE=\"1\">\\n\');\
</SCRIPT>\n");

   RequestPublicStateSave (rdptr);

   fprintf (stdout,
"</FORM>\n\
<!-- %s -->\n\
</BODY>\n\
</HTML>\n",
           StatTimer(TRUE));
}

/*****************************************************************************/
/*
Equvalent to the main menu panel in private access soyMAIL.
*/

void PublicPageTitle (REQUEST_DATA *rdptr)

{
   char  *cptr,
         *Title1Ptr,
         *Title2Ptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("PublicPageTitle()");

   fprintf (stdout,
"<!-- public title begin -->\n\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"0\" \
ALIGN=\"center\" CLASS=\"css_panel_1\">\n\
<TR><TD>\n\
<TABLE CLASS=\"css_panel_2\" WIDTH=\"100%%\" ALIGN=\"center\" \
CELLPADDING=\"4\" CELLSPACING=\"2\">\n");

   fprintf (stdout,
"<TR>\n\
<TD NOWRAP ALIGN=\"left\" VALIGN=\"middle\" WIDTH=\"10%%\">&nbsp;&nbsp;");

   if (!strcmp (SOY_MAIL, "soyMAIL"))
      fprintf (stdout,
"<SPAN STYLE=\"font-weight:bold;font-size:120%%;letter-spacing:1px;\">so</SPAN>\
<SPAN STYLE=\"font-weight:bold;font-size:120%%;letter-spacing:2px;\">yMAIL</SPAN>");
   else
      fprintf (stdout,
"<SPAN STYLE=\"font-weight:bold;font-size:120%%;letter-spacing:2px;\">%s</SPAN>\n",
               SOY_MAIL);

   fprintf (stdout, "&nbsp;&nbsp;</TD>\n");

   if (rdptr->PublicWildcard || SoyMailConfig.SearchEnabled)
      fprintf (stdout, "<TD NOWRAP ALIGN=\"left\" VALIGN=\"middle\">\n\n");

   if (rdptr->PublicWildcard)
   {
      FolderListInit (rdptr);
      FolderOpenSelect (rdptr);
   }

   if (SoyMailConfig.SearchEnabled)
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"main_btn\" VALUE=\"%s\">&nbsp;\n",
              LangFor("search"));

   fprintf (stdout,
"<SCRIPT LANGUAGE=\"JavaScript\">\
if (window.opener) \
document.write(\'<INPUT CLASS=\"css_button_panel_2\" \
TYPE=\"submit\" NAME=\"main_btn\" VALUE=\"%s\" \
ONCLICK=\"window.close()\">&nbsp;\\n\');\
</SCRIPT>\n",
           LangFor("close"));

   if (rdptr->PublicWildcard || SoyMailConfig.SearchEnabled)
      fprintf (stdout, "</TD>\n");

   Title2Ptr = NULL;
   if (SoyMailConfig.PageTitle1)
   {
      Title1Ptr = SoyMailConfig.PageTitle1;
      if (rdptr->PublicAccessTitle[0])
         Title2Ptr = rdptr->PublicAccessTitle;
      else
      if (SoyMailConfig.PageTitle2)
         Title2Ptr = SoyMailConfig.PageTitle2;
   }
   else
   if (rdptr->PublicAccessTitle[0])
      Title1Ptr = rdptr->PublicAccessTitle;
   else
      Title1Ptr = rdptr->CgiPathInfoPtr;

   if (Title1Ptr && Title2Ptr)
      fprintf (stdout,
"<TD ALIGN=\"middle\" WIDTH=\"90%%\" \
STYLE=\"font-weight:bold;font-size:120%%;\">%s&nbsp;</TD>\
<TD ALIGN=\"right\" WIDTH=\"10%%\" \
STYLE=\"font-weight:bold;font-size:120%%;\">%s&nbsp;</TD>",
               Title1Ptr,
               Title2Ptr);
   else
      fprintf (stdout,
"<TD ALIGN=\"right\" WIDTH=\"100%%\" \
STYLE=\"font-weight:bold;font-size:120%%;\">%s&nbsp;</TD>",
               Title1Ptr);

   fprintf (stdout,
"\n\
</TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
<!-- public title end -->\n\n");

}

/*****************************************************************************/
/*
Display the basic message folder and message list page.
*/

void PublicMessageFolderList (REQUEST_DATA *rdptr)

{
   int  cnt, idx, status;

   char  *cptr, *sptr, *zptr;
   char  FromBuffer [256],
         MessageReadAgo [32],
         SubjBuffer [256];
   USER_OPTIONS  *uoptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("PublicMessageFolderList()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;
   vmptr = &muptr->VmsMailMsg;

   fprintf (stdout,
"<P CLASS=\"css_p08\">\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"5\" \
ALIGN=\"center\" CLASS=\"css_panel_1\">\n\
<TR>\n");

   /******************/
   /* message ranges */
   /******************/

   MessageRangeSelect (rdptr);

   if (!vmptr->MessageSelectedCount)
   {
      /****************/
      /* empty folder */
      /****************/

      rdptr->RangeIdStart = rdptr->RangeIdCount = 0;

      fprintf (stdout, "<TD></TD></TR></TABLE>\n");

      MessageListFooter ();

      fprintf (stdout,
"</TD></TR>\n\
</TABLE>\n\
<!-- public message list end -->\n\n");

      return;
   }

   /********************/
   /* list of messages */
   /********************/

   fprintf (stdout,
"<TR><TD COLSPAN=\"2\">\n\
<TABLE CELLPADDING=\"0\" CELLSPACING=\"5\" BORDER=\"0\">\n\
<TR CLASS=\"css_bu\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
function targetPopup (pop) {\n\
   document.soymail_form.msg_btn_popup.value = pop;\n\
   if (pop) targetBlank();\n\
}\n\
</SCRIPT>\n\
<INPUT TYPE=\"hidden\" NAME=\"msg_btn_popup\" VALUE=\"\">\n\
<INPUT TYPE=\"hidden\" NAME=\"msg_range_list\" VALUE=\"%s\">\n\
<TD></TD><TD></TD>\
<TD CLASS=\"css_nwtar5\">%s</TD>\
<TD CLASS=\"css_nwtal5\">%s</TD>\
<TD CLASS=\"css_nwtal5\">%s</TD>\
</TR>\n",
            rdptr->MsgIdListPtr,
            LangFor("received"),
            LangFor("from"),
            LangFor("subject"));

   for (vmptr = muptr->MessageList; vmptr; vmptr = vmptr->NextPtr)
   {
      if (Debug)
         fprintf (stdout,
"%%X%08.08X id:%d size:%d hash:%d\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n\n",
                  muptr->MailVmsStatus,
                  vmptr->CurrentId, vmptr->SizeInRecords, vmptr->MessageHash,
                  vmptr->Date, vmptr->From, vmptr->Subject,
                  vmptr->To, vmptr->Cc);

      zptr = (sptr = FromBuffer) + sizeof(FromBuffer)-1;
      for (cptr = vmptr->From; *cptr && sptr < zptr; *sptr++ = *cptr++);
      *sptr = '\0';
      MimeDecIsoString (FromBuffer);
      FromBuffer[40] = ']';
      FromBuffer[41] = '\0';

      zptr = (sptr = SubjBuffer) + sizeof(SubjBuffer)-1;
      for (cptr = vmptr->Subject; *cptr && sptr < zptr; *sptr++ = *cptr++);
      *sptr = '\0';
      MimeDecIsoString (SubjBuffer);
      SubjBuffer[60] = ']';
      SubjBuffer[61] = '\0';

      /* indicate whether this message recently has been read */
      strcpy (MessageReadAgo, "<TD></TD>");
      if (uoptr->MessageReadHistory)
      {
         for (idx = 0; idx < rdptr->MessageReadHistoryCount; idx++)
         {
            if (rdptr->MessageReadHistory[idx] != vmptr->CurrentId) continue;
            if (uoptr->MessageReadHistory == OPTIONS_READ_HISTORY_ASTER)
               strcpy (MessageReadAgo, "<TD CLASS=\"css_rhist\">*</TD>");
            else
               sprintf (MessageReadAgo, "<TD CLASS=\"css_rhist\">-%d</TD>", idx+1);
            break;
         }
      }

      /* this mix of styles to get both Mozilla and MSIE to look right! */
      fprintf (stdout,
"<TR CLASS=\"css_nw\">\
<TD>\
<INPUT CLASS=\"css_button_panel_1\" \
TYPE=\"submit\" NAME=\"msg_btn\" VALUE=\"%04d\" \
ONCLICK=\"targetPopup(0);\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_btn_%d\" VALUE=\"^\" \
ONCLICK=\"targetPopup(%d);\">\');\
</SCRIPT>\
</TD>\
%s\
<TD CLASS=\"css_nwtar5\">%17.17s</TD>\
<TD CLASS=\"css_nwtal5\">%s</TD>\
<TD CLASS=\"css_nwtal5\">%s</TD>\
</TR>\n",
         vmptr->CurrentId,
         vmptr->CurrentId,
         vmptr->CurrentId,
         MessageReadAgo,
         InetMailRfcVmsDate(vmptr->Date),
         HTML_ADDR(FromBuffer),
        vmptr->Subject[0] ? HTML_TEXT(SubjBuffer) : LangFor("(none)"));
   }

   fprintf (stdout,
"</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
<!-- public message list end -->\n\n");

   MessageListFooter ();
}

/*****************************************************************************/
/*
Display the basic public message list page.
*/

void PublicFolderPage (REQUEST_DATA *rdptr)

{
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("PublicFolderPage()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;
   vmptr = &muptr->VmsMailMsg;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;

   if (!rdptr->RangeIdStart)
   {
      rdptr->RangeIdStart = sdptr->RangeIdStart;
      rdptr->RangeIdCount = sdptr->RangeIdCount;
      if (!rdptr->RangeIdStart)
      {
         /* no previous range to fall back on */
         rdptr->RangeIdStart = 1;
         rdptr->RangeIdCount = RANGE_MOST_RECENT;
      }
   }
   if (!rdptr->RangeIdCount)
      rdptr->RangeIdCount = RANGE_MOST_RECENT;

   muptr->MailVmsStatus = SS$_NORMAL;

   MessageListInit (rdptr);

   PublicPageBegin (rdptr, PAGE_FOLDER);

   PublicPageTitle (rdptr);

   StatusInfoPanel (rdptr);

   PublicMessageFolderList (rdptr);

   PublicPageEnd (rdptr);
}

/*****************************************************************************/
/*
Display the public format message header and body page.
*/

void PublicMessagePage (REQUEST_DATA *rdptr)

{
   int  status;
   REQUEST_DATA  *sdptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;
   VMS_MAIL_USER  UserData;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("PublicMessagePage()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;
   vmptr = &muptr->VmsMailMsg;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;

   rdptr->FolderMessageCount = sdptr->FolderMessageCount;

   vmptr->MessageId = rdptr->MessageId;
   vmptr->ConfirmHash = rdptr->MessageHash;

   CallMailMessageGet (muptr, NULL);

   /* in case a message hash was not supplied use the one just generated */
   rdptr->MessageHash = vmptr->MessageHash;

   if (VMSnok (muptr->MailVmsStatus))
   {
      if (vmptr->ErrorOpenIn)
      {
         /* error opening external message file */
         StatusMessage (FI_LI, 1, "MAIL: %s %s",
                        SysGetMsg(MAIL$_OPENIN), vmptr->ExtId);
      }
      else
      {
         if (muptr->MailVmsStatus == SS$_BADCHECKSUM ||
             muptr->MailVmsStatus == MAIL$_NOMOREMSG)
            StatusMessage (FI_LI, 1, "%s", LangFor("message_checksum"));
         else
            StatusMessage (FI_LI, 1, "MAIL: %s.",
                           SysGetMsg(muptr->MailVmsStatus));
      }
      PublicFolderPage (rdptr);
      return;
   }

   MessageContentIdentify (rdptr);

   PublicPageBegin (rdptr, PAGE_MESSAGE);

   PublicPageTitle (rdptr);

   StatusInfoPanel (rdptr);

   PublicMessageButtons (rdptr);

   MessageHeader (rdptr);

   MessageBody (rdptr);

   MessageAttachments (rdptr);

   PublicPageEnd (rdptr);
}

/*****************************************************************************/
/*
Display the buttons associated with the message.
*/

void PublicMessageButtons (REQUEST_DATA *rdptr)

{
   USER_OPTIONS  *uoptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("PublicMessageButtons()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;
   vmptr = &muptr->VmsMailMsg;

   fprintf (stdout,
"<!-- header buttons begin -->\n\
<P CLASS=\"css_p08\">\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"5\" \
ALIGN=\"center\" CLASS=\"css_panel_1\">\n\
<TR><TD STYLE=\"text-align:left;\">\n\
\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_print\" VALUE=\"%s\">\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_print_popup\" VALUE=\"^\" \
ONCLICK=\"targetBlank();\">\');\
</SCRIPT> \n\
\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_raw\" VALUE=\"%s\"%s>\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_raw_popup\" VALUE=\"^\"%s>\');\
</SCRIPT> \n\
\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_hdr\" VALUE=\"%s\"%s>\
<SCRIPT LANGUAGE=\"JavaScript\">\
document.write(\'<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"border-left-width:0px;\" \
TYPE=\"submit\" NAME=\"msg_hdr_popup\" VALUE=\"^\"%s>\');\
</SCRIPT> \n\
\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_prev\" VALUE=\"%s\"%s> \n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_folder_at\" VALUE=\"%s\"%s> \n\
<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_next\" VALUE=\"%s\"%s> \n",
            LangFor("print"),
            LangFor("raw"),
            vmptr->LooksLikeRfc ? "" : "ONCLICK=\"return notApplicable();\"",
            vmptr->LooksLikeRfc ? "ONCLICK=\"targetBlank();\"" :
                                  "ONCLICK=\"return notApplicable();\"",
            LangFor("header"),
            vmptr->LooksLikeRfc ? "" : "ONCLICK=\"return notApplicable();\"",
            vmptr->LooksLikeRfc ? "ONCLICK=\"targetBlank();\"" :
                                  "ONCLICK=\"return notApplicable();\"",
            LangFor("prev"),
            CLICK_NOMORE(vmptr->MessageId == 1),
            LangFor("folder_at"),
            CLICK_STARTWORKING,
            LangFor("next"),
            CLICK_NOMORE(vmptr->MessageId == rdptr->FolderMessageCount));

   if (rdptr->ReadPlainText && rdptr->HtmlTextIsPresent)
   {
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_read_pref\" VALUE=\"%s\"> \n",
               LangFor("message_read_html"));
   }
   else
   if (rdptr->ReadHtmlText && rdptr->PlainTextIsPresent)
   {
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\"  \
NAME=\"msg_read_pref\" VALUE=\"%s\"> \n",
               LangFor("message_read_plain"));
   }

   if (rdptr->ReadPlainText)
   {
      /* one is on, minus one is off, zero is not set! */
      if (rdptr->MassageMessage)
         fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_wrap\" VALUE=\"%s\">\n", LangFor("message_massage_off"));
      else
         fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" TYPE=\"submit\" \
NAME=\"msg_wrap\" VALUE=\"%s\">\n", LangFor("message_massage_on"));
   }

   fprintf (stdout,
"</TD></TR>\n\
</TABLE>\n\
<!-- header buttons end -->\n\n");

}

/****************************************************************************/

